#include <time.h>
#include <glob.h>
#include <unistd.h>
#include <string>
#include <list>
#include <sstream>
#include <iomanip>
#include <iostream>
#include <gtkmm/messagedialog.h>
#include <gtkmm/main.h>
#include "DataManagerDialog.h"
#include "DataStore.h"
#include "ApproachPlate.h"
#include "Util.h"
#include "ImportDialog.h"

DataManagerDialog::DataManagerDialog(Config *Cfg, std::string *GladeFilename, DataStore *Data, Shapefile *ShapefileData)
{
	try
	{
		refXml = Gnome::Glade::Xml::create(*GladeFilename, "DataManagerDialog");
		ProgRefXml = Gnome::Glade::Xml::create(*GladeFilename, "ProgressWindow");
	}
	catch (const Gnome::Glade::XmlError& ex)
	{
		Gtk::MessageDialog *message = new Gtk::MessageDialog("Error reading data file.  Program cannot continue.",
			false,Gtk::MESSAGE_ERROR,Gtk::BUTTONS_CANCEL,true);
		message->run();
		Gtk::Main::quit();
	}

	refXml->get_widget("DataManagerDialog",GtkDialog);
	refXml->get_widget("ActivateButton",ActivateButton);
	refXml->get_widget("DeleteButton",DeleteButton);
	refXml->get_widget("ImportButton",ImportButton);
	refXml->get_widget("DataManagerTreeView",DataManagerTreeView);

	ProgRefXml->get_widget("ProgressWindow",ProgressWindow);
	ProgRefXml->get_widget("SubProgressBar",SubProgressBar);
	ProgRefXml->get_widget("ProgressBar",ProgressBar);
	ProgRefXml->get_widget("CancelButton",CancelButton);

	ProgressWindow->set_deletable(false);
	CancelButton->signal_clicked().connect(sigc::mem_fun(this,&DataManagerDialog::on_CancelButton_clicked));

	ActivateButton->signal_clicked().connect(sigc::mem_fun(this,&DataManagerDialog::on_ActivateButton_clicked));
	DeleteButton->signal_clicked().connect(sigc::mem_fun(this,&DataManagerDialog::on_DeleteButton_clicked));
	ImportButton->signal_clicked().connect(sigc::mem_fun(this,&DataManagerDialog::on_ImportButton_clicked));

	this->Cfg=Cfg;
	this->GladeFilename=GladeFilename;
	this->Data=Data;
	this->ShapefileData=ShapefileData;

	ListStore = Gtk::ListStore::create(Columns);
	DataManagerTreeView->set_model(ListStore);
	DataManagerTreeView->append_column("Active", Columns.Active);
	DataManagerTreeView->append_column("Volume", Columns.VolumeName);
	DataManagerTreeView->append_column("Eff. From", Columns.EffFrom);
	DataManagerTreeView->append_column("Eff. To", Columns.EffTo);
	DataManagerTreeView->append_column("Airports", Columns.nAirports);
	DataManagerTreeView->append_column("NavAids", Columns.nNavAids);
	DataManagerTreeView->append_column("Waypoints", Columns.nWaypoints);
	DataManagerTreeView->append_column("Charts", Columns.nCharts);
	DataManagerTreeView->append_column("Approach Plates", Columns.nApproachPlates);

	/* Initial sort */
	ListStore->set_sort_column(Columns.VolumeName, Gtk::SORT_DESCENDING);

	/* Allow re-sorting by any column */
	Gtk::TreeView::Column* pColumn = DataManagerTreeView->get_column(0);
	if(pColumn)
		pColumn->set_sort_column(Columns.Active);

	pColumn = DataManagerTreeView->get_column(1);
	if(pColumn)
		pColumn->set_sort_column(Columns.VolumeName);

	pColumn = DataManagerTreeView->get_column(2);
	if(pColumn)
		pColumn->set_sort_column(Columns.EffFrom);

	pColumn = DataManagerTreeView->get_column(3);
	if(pColumn)
		pColumn->set_sort_column(Columns.EffTo);

	pColumn = DataManagerTreeView->get_column(4);
	if(pColumn)
		pColumn->set_sort_column(Columns.nAirports);

	pColumn = DataManagerTreeView->get_column(5);
	if(pColumn)
		pColumn->set_sort_column(Columns.nNavAids);

	pColumn = DataManagerTreeView->get_column(6);
	if(pColumn)
		pColumn->set_sort_column(Columns.nWaypoints);

	pColumn = DataManagerTreeView->get_column(7);
	if(pColumn)
		pColumn->set_sort_column(Columns.nCharts);

	pColumn = DataManagerTreeView->get_column(8);
	if(pColumn)
		pColumn->set_sort_column(Columns.nApproachPlates);

	PopulateTreeView();
	CancelFlag=false;
}

DataManagerDialog::~DataManagerDialog()
{
	delete GtkDialog;
	delete ProgressWindow;
	Data->ClearUnloadedVolumes();
}

void DataManagerDialog::Run() {
	GtkDialog->run();
}

void DataManagerDialog::on_ActivateButton_clicked() {
	Glib::RefPtr<Gtk::TreeSelection> DataManagerSelection = DataManagerTreeView->get_selection();
	Gtk::TreeModel::iterator iter = DataManagerSelection->get_selected();
	if (iter) {		/* If anything is selected */
		Gtk::TreeModel::Row row = *iter;

		if (Data->SetActiveVolume(row[Columns.Volumei])) {
			/* Update the display to indicate the new active volume */
			for (iter=DataManagerTreeView->get_model()->get_iter("0");iter;++iter) {
				row=*iter;
				if (((std::list<Volume*>::iterator)row[Columns.Volumei])==Data->GetActiveVolume()) {
					row[Columns.Active]="*";
				} else {
					row[Columns.Active]="";
				}
			}
		}
	}
}

void DataManagerDialog::on_DeleteButton_clicked() {
	std::ostringstream Text;
	Gtk::MessageDialog *Message;

	Glib::RefPtr<Gtk::TreeSelection> DataManagerSelection = DataManagerTreeView->get_selection();
	Gtk::TreeModel::iterator iter = DataManagerSelection->get_selected();
	if (iter) {		/* If anything is selected */
		Gtk::TreeModel::Row row = *iter;
		std::list<Volume*>::iterator Volumei=row[Columns.Volumei];

		Text<<"Are you sure you want to delete Volume "<<(*Volumei)->GetName()<<"?";
		Message = new Gtk::MessageDialog(Text.str(),
			false,Gtk::MESSAGE_QUESTION,Gtk::BUTTONS_OK_CANCEL,true);
		if (Message->run()!=Gtk::RESPONSE_OK) {
			delete Message;
			return;
		}
		delete Message;

		(*Volumei)->Load();

		DeleteVolume(Volumei);

		Data->GetVolumes()->remove(*Volumei);
		delete *Volumei;
		ListStore->erase(row);

		if (Data->GetActiveVolume()==Volumei) {
			/* If we're deleting the currently active volume, reset
			 * the active volume to either the current volume or
			 * nothing.
			 */
			Data->ActivateCurrentVolume();
		}
	} else {
		Message = new Gtk::MessageDialog("No data volume selected",
			false,Gtk::MESSAGE_ERROR,Gtk::BUTTONS_CANCEL,true);
		Message->run();
		delete Message;
		return;
	}
}

void DataManagerDialog::on_ImportButton_clicked() {
	ImportDialog *Dialog=new ImportDialog(Cfg,GladeFilename,Data,ShapefileData);
	Dialog->Run();
	delete Dialog;
	PopulateTreeView();	/* Anything could have changed, so reset everything */
}

void DataManagerDialog::on_CancelButton_clicked() {
	Gtk::MessageDialog *Message = new Gtk::MessageDialog(
		"Are you sure you want to cancel?  This will leave useless files on your disk.",
		false,Gtk::MESSAGE_QUESTION,Gtk::BUTTONS_YES_NO,true);
	if (Message->run()!=Gtk::RESPONSE_YES) {
		CancelFlag=true;
	}
	delete Message;
}

void DataManagerDialog::PopulateTreeView() {
	time_t EffTime;
	struct tm bdtime;
	std::ostringstream TimeStr;

	ListStore->clear();
	for (std::list<Volume*>::iterator Volumei=Data->GetVolumes()->begin();Volumei!=Data->GetVolumes()->end();++Volumei) {
		Gtk::TreeModel::iterator iter = ListStore->append();
		Gtk::TreeModel::Row row = *iter;

		if (Volumei==Data->GetActiveVolume())
			row[Columns.Active] = "*";
		else
			row[Columns.Active] = "";
		row[Columns.VolumeName] = (*Volumei)->GetName();

		TimeStr.seekp(0);
		EffTime=(*Volumei)->GetEffectiveFrom();
		gmtime_r(&EffTime,&bdtime);
		TimeStr<<std::setw(4)<<std::setfill('0')<<(bdtime.tm_year+1900)<<"-"<<std::setw(2)<<(bdtime.tm_mon+1)<<"-"<<std::setw(2)<<(bdtime.tm_mday);
		row[Columns.EffFrom] = TimeStr.str();

		TimeStr.seekp(0);
		EffTime=(*Volumei)->GetEffectiveTo();
		gmtime_r(&EffTime,&bdtime);
		TimeStr<<std::setw(4)<<std::setfill('0')<<(bdtime.tm_year+1900)<<"-"<<std::setw(2)<<(bdtime.tm_mon+1)<<"-"<<std::setw(2)<<(bdtime.tm_mday);
		row[Columns.EffTo] = TimeStr.str();

		row[Columns.nAirports] = (*Volumei)->GetnAirports();
		row[Columns.nNavAids] = (*Volumei)->GetnNavAids();
		row[Columns.nWaypoints] = (*Volumei)->GetnEnrouteWaypoints();
		row[Columns.nCharts] = (*Volumei)->GetnGeoCharts();
		row[Columns.nApproachPlates] = (*Volumei)->GetnApproachPlates();
		row[Columns.Volumei] = Volumei;
	}

	/* Since the size of the data isn't known until runtime, recenter the dialog here */
	GtkDialog->set_position(Gtk::WIN_POS_CENTER_ALWAYS);
}

bool DataManagerDialog::DeleteVolume(std::list<Volume*>::iterator Volumei) {
	std::string Dir;

	ProgressBar->set_text("Total");
	SubProgressBar->set_text("");
	SubProgressBar->show();
	ProgressWindow->show();
	Dir=(*Volumei)->GetDirectory();

	DeleteGeoCharts(Dir+"/geocharts",(*Volumei)->GetGeoCharts());
	DeleteApproachPlates(Dir+"/plateimages",(*Volumei)->GetAirports());

	/* Remove DataStore file */
	unlink((Dir+"/volume.vifir").c_str());

	/* Remove Directory */
	if (rmdir(Dir.c_str())!=0) {
		Gtk::MessageDialog *Message = new Gtk::MessageDialog("Unable to completely remove "+Dir+".  It may be removed manually.",
			false,Gtk::MESSAGE_WARNING,Gtk::BUTTONS_OK,true);
		Message->run();
		delete Message;
	}

	ProgressWindow->hide();
	return true;
}

bool DataManagerDialog::DeleteGeoCharts(std::string Dir,std::vector<GeoChart*> *GeoCharts) {
	std::string Path;
	bool Errors=false;
	glob_t globdata;
	unsigned int i=0;

	SubProgressBar->set_text("Charts");

	/* Remove known GeoCharts */
	for (std::vector<GeoChart*>::iterator Charti=GeoCharts->begin();Charti!=GeoCharts->end() && !CancelFlag;++Charti) {
		Path=Dir+"/"+(*Charti)->GetFileName();
		if (unlink(Path.c_str())!=0) {
			Errors=true;
		}
		i++;
		SubProgressBar->set_fraction((double)i/GeoCharts->size());
		ProgressBar->set_fraction(((double)i/GeoCharts->size())/2.0);
		Util::ScheduleMainLoop();
	}

	/* Remove any other files in GeoCharts dir */
	if (glob((Dir+"/*").c_str(),0,NULL,&globdata)) {
		for (i=0;i<globdata.gl_pathc;i++) {
			if (unlink(globdata.gl_pathv[i])!=0) {
				Errors=true;
			}
			SubProgressBar->pulse();
			Util::ScheduleMainLoop();
		}
		globfree(&globdata);
	}

	if (rmdir(Dir.c_str())!=0)
		Errors=true;

	return !Errors;
}

bool DataManagerDialog::DeleteApproachPlates(std::string Dir,std::vector<Airport*> *Airports) {
	std::string Path;
	bool Errors=false;
	glob_t globdata;
	unsigned int i=0;

	SubProgressBar->set_text("Approach Plates");

	/* Remove front matter */
	if (unlink((Dir+"/"+FRONTMATTERFILENAME).c_str())!=0) {
		Errors=true;
	}

	/* Remove knwon Approaches */
	for (std::vector<Airport*>::iterator Airporti=Airports->begin();Airporti!=Airports->end() && !CancelFlag;++Airporti) {
		for (std::vector<ApproachPlate*>::iterator ApproachPlatei=(*Airporti)->GetApproachPlates()->begin();ApproachPlatei!=(*Airporti)->GetApproachPlates()->end() && !CancelFlag;++ApproachPlatei) {
			Path=Dir+"/"+(*ApproachPlatei)->GetImgName();
			if (unlink(Path.c_str())!=0) {
				Errors=true;
			}
		}
		i++;
		SubProgressBar->set_fraction((double)i/Airports->size());
		ProgressBar->set_fraction(((double)i/Airports->size())/2.0+0.5);
		Util::ScheduleMainLoop();
	}

	/* Remove any other files in Approach Plates dir */
	if (glob((Dir+"/*").c_str(),0,NULL,&globdata)) {
		for (i=0;i<globdata.gl_pathc;i++) {
			if (unlink(globdata.gl_pathv[i])!=0) {
				Errors=true;
			}
			SubProgressBar->pulse();
			Util::ScheduleMainLoop();
		}
		globfree(&globdata);
	}

	if (rmdir(Dir.c_str())!=0)
		Errors=true;

	return !Errors;
}

DataManagerDialogModelColumns::DataManagerDialogModelColumns() {
	add(Active);
	add(VolumeName);
	add(EffFrom);
	add(EffTo);
	add(nAirports);
	add(nNavAids);
	add(nWaypoints);
	add(nCharts);
	add(nApproachPlates);
	add(Volumei);
}
