#ifndef DATASTORE_H_
#define DATASTORE_H_

#include <stdint.h>
#include <time.h>
#include <string>
#include <vector>
#include <list>
#include "Airport.h"
#include "Runway.h"
#include "Frequency.h"
#include "NavAid.h"
#include "EnrouteWaypoint.h"
#include "ApproachPlate.h"
#include "GeoChart.h"
#include "CRC32.h"
#include "Config.h"

#define VOLUME_FILEMAGIC "Vifir_Volume"
#define VOLUME_MAXSEARCHRESULTS 256

class Volume
{
public:
	enum RecordType_enum {
		RECORDTYPE_AIRPORT,
		RECORDTYPE_RUNWAY,
		RECORDTYPE_FREQ,
		RECORDTYPE_NAVAID,
		RECORDTYPE_ENROUTEWAYPOINT,
		RECORDTYPE_APPROACHPLATE,
		RECORDTYPE_GEOCHART
	};

	Volume(std::string *GladeFilename, Config *Cfg);
	Volume(std::string *GladeFilename, Config *Cfg, std::string Name, time_t EffectiveFrom, time_t EffectiveTo);
	virtual ~Volume();

	bool Load(std::string Filename="");
	void Save(std::string Filename="");
	void SaveAirports(std::vector<Airport*> *AirportV,std::ofstream *ofs, uint32_t *RunningSize, uint32_t *RunningCRC);
	void SaveRunways(std::vector<Runway*> *RunwayV,std::ofstream *ofs, uint32_t *RunningSize, uint32_t *RunningCRC);
	void SaveFreqs(std::vector<Frequency*> *FreqV,std::ofstream *ofs, uint32_t *RunningSize, uint32_t *RunningCRC);
	void SaveNavAids(std::vector<NavAid*> *NavAidV,std::ofstream *ofs, uint32_t *RunningSize, uint32_t *RunningCRC);
	void SaveEnrouteWaypoints(std::vector<EnrouteWaypoint*> *EnrouteWaypointV,std::ofstream *ofs, uint32_t *RunningSize, uint32_t *RunningCRC);
	void SaveApproachPlates(std::vector<ApproachPlate*> *ApproachPlateV,std::ofstream *ofs, uint32_t *RunningSize, uint32_t *RunningCRC);
	void SaveGeoCharts(std::vector<GeoChart*> *GeoChartV,std::ofstream *ofs, uint32_t *RunningSize, uint32_t *RunningCRC);
	void LoadAirport(std::vector<Airport*> *AirportV,std::ifstream *ifs, uint32_t *RunningCRC);
	void LoadRunway(std::vector<Runway*> *RunwayV,std::ifstream *ifs, uint32_t *RunningCRC);
	void LoadFreq(std::vector<Frequency*> *FreqV,std::ifstream *ifs, uint32_t *RunningCRC);
	void LoadNavAid(std::vector<NavAid*> *NavAidV,std::ifstream *ifs, uint32_t *RunningCRC);
	void LoadEnrouteWaypoint(std::vector<EnrouteWaypoint*> *EnrouteWaypointV,std::ifstream *ifs, uint32_t *RunningCRC);
	void LoadApproachPlate(std::vector<ApproachPlate*> *ApproachPlateV,Airport *AP,std::ifstream *ifs, uint32_t *RunningCRC);
	void LoadGeoChart(std::vector<GeoChart*> *GeoChartV,std::ifstream *ifs, uint32_t *RunningCRC);
	int RecordSize(struct Volume_Record_struct *Record);
	void Clear();
	void ClearFlightDB();
	void ClearGeoCharts();
	void ClearApproachPlates();
	void UpdateCounts();
	std::string GetDataDir();
	bool Scan(std::string Filename="", struct Volume_Header_struct *Header=NULL);
	static time_t ComputeEffectiveFrom(std::string VolStr);
	static time_t ComputeEffectiveTo(std::string VolStr);

	Airport *GetAirportByICAO(std::string ICAO);
	Airport *GetAirportByIATA(std::string IATA);
	std::vector<Airport*> *GetAirports();
	unsigned int GetnAirports();
	std::vector<NavAid*> *GetNavAids();
	unsigned int GetnNavAids();
	std::vector<EnrouteWaypoint*> *GetEnrouteWaypoints();
	unsigned int GetnEnrouteWaypoints();
	std::vector<GeoChart*> *GetGeoCharts();
	unsigned int GetnGeoCharts();
	unsigned int GetnApproachPlates();
	void WaypointSearch(std::string Str, std::vector<Waypoint*> *Waypoints, unsigned int Max=VOLUME_MAXSEARCHRESULTS);
	std::string GetName();
	time_t GetEffectiveFrom();
	time_t GetEffectiveTo();
	void GetSectionalsByBounds(double Lon, double Lat, std::vector<GeoChart*> *MatchingSectionals);
	GeoChart* GetBestSectional(double Lon, double Lat, std::vector<GeoChart*> *Choices=NULL);
	bool GetLoaded();
	std::string GetDirectory();
	bool IsPlateImageFileUsed(std::string Filename);

private:
	static const double ReadingProgressFrac=0.2;
	static const double AirportsProgressFrac=0.2;
	static const double NavAidsProgressFrac=0.2;
	static const double WaypointsProgressFrac=0.2;
	static const double ChartsProgressFrac=0.2;

	std::string *GladeFilename;
	Config *Cfg;

	std::string Filename;
	std::string Name;
	time_t EffectiveFrom,EffectiveTo;
	std::vector<Airport*> Airports;
	std::vector<NavAid*> NavAids;
	std::vector<EnrouteWaypoint*> EnrouteWaypoints;
	std::vector<GeoChart*> GeoCharts;
	CRC32 CRC;
	bool CancelFlag;
	bool Loaded;
	/* Keep track of the data counts even when the data isn't loaded */
	unsigned int nAirports, nNavAids, nEnrouteWaypoints, nGeoCharts, nApproachPlates;

	void on_CancelButton_clicked();
};

/* Increment the version when the data structure changes */
#define VOLUME_HEADER_STRUCT_VER 1
struct Volume_Header_struct {
	char Magic[sizeof(VOLUME_FILEMAGIC)];
	uint32_t DataVersion;
	char VolumeName[4];
	uint32_t EffectiveFrom,EffectiveTo;
	uint32_t FileSize;
	uint32_t CRC32;	/* Computed with this field and FileSize set to zero */
	uint32_t nAirports;
	uint32_t nNavAids;
	uint32_t nEnrouteWaypoints;
	uint32_t nGeoCharts;
	uint32_t nApproachPlates;
} __attribute__ ((packed));

/* Increment VOLUME_HEADER_STRUCT_VER when this data structure changes */
struct Volume_Record_struct {
	uint8_t Type;
	union {
		struct Airport_Binary_struct Airport;
		struct Runway_Binary_struct Runway;
		struct Frequency_Binary_struct Freq;
		struct NavAid_Binary_struct NavAid;
		struct EnrouteWaypoint_Binary_struct EnrouteWaypoint;
		struct ApproachPlate_Binary_struct ApproachPlate;
		struct GeoChart_Binary_struct GeoChart;
	};
} __attribute__ ((packed));

class DataStore {
public:
	DataStore(std::string *GladeFilename, Config *Cfg);
	virtual ~DataStore();

	Airport *GetAirportByICAO(std::string ICAO);
	Airport *GetAirportByIATA(std::string IATA);
	std::vector<Airport*> *GetAirports();
	unsigned int GetnAirports();
	std::vector<NavAid*> *GetNavAids();
	unsigned int GetnNavAids();
	std::vector<EnrouteWaypoint*> *GetEnrouteWaypoints();
	unsigned int GetnEnrouteWaypoints();
	std::vector<GeoChart*> *GetGeoCharts();
	unsigned int GetnGeoCharts();
	unsigned int GetnApproachPlates();
	void WaypointSearch(std::string Str, std::vector<Waypoint*> *Waypoints);
	time_t GetEffectiveFrom();
	time_t GetEffectiveTo();
	void GetSectionalsByBounds(double Lon, double Lat, std::vector<GeoChart*> *MatchingSectionals);
	GeoChart* GetBestSectional(double Lon, double Lat, std::vector<GeoChart*> *Choices=NULL);
	bool HasActiveVolume();
	std::list<Volume*>::iterator GetActiveVolume();
	bool SetActiveVolume(std::list<Volume *>::iterator Vol);
	Volume *GetVolume(std::string *Name, time_t *EffectiveFrom, time_t *EffectiveTo);
	void Clear();
	void ClearVolumes();
	void ClearUnloadedVolumes();
	void ScanVolumes();
	void ActivateCurrentVolume();
	std::list<Volume*> *GetVolumes();

private:
	std::string *GladeFilename;
	Config *Cfg;
	std::list<Volume*> Volumes;	/* Use a list so the ActiveVolume is not invalidated when a new volume is added */
	std::list<Volume*>::iterator ActiveVolume;
};

#endif /*DATASTORE_H_*/
