#include <glob.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <gtkmm/main.h>
#include <gtkmm/stock.h>
#include <gtkmm/messagedialog.h>
#include <gtkmm/filechooserdialog.h>
#include <libglademm.h>
#include "ImportSectionalsDVD.h"
#include "Config.h"
#include "SectionalChart.h"
#include "Exception.h"
#include "Util.h"
#include "GeoChart.h"
#include "ImportSectionalsDVDVolumeChooserDialog.h"

ImportSectionalsDVD::ImportSectionalsDVD(Config *Cfg, std::string *GladeFilename, DataStore *Data)
{
	this->Cfg=Cfg;
	this->GladeFilename=GladeFilename;
	this->Data=Data;

	try
	{
		std::string gladefile;

		refXml = Gnome::Glade::Xml::create(*GladeFilename, "ProgressWindow");
	}
	catch (const Gnome::Glade::XmlError& ex)
	{
		Gtk::MessageDialog *message = new Gtk::MessageDialog("Error reading data file.  Program cannot continue.",
			false,Gtk::MESSAGE_ERROR,Gtk::BUTTONS_CANCEL,true);
		message->run();
		Gtk::Main::quit();
	}

	refXml->get_widget("ProgressWindow",ProgressWindow);
	refXml->get_widget("SubProgressBar",SubProgressBar);
	refXml->get_widget("ProgressBar",ProgressBar);
	refXml->get_widget("CancelButton",CancelButton);

	ProgressWindow->set_deletable(false);
	CancelButton->signal_clicked().connect(sigc::mem_fun(this,&ImportSectionalsDVD::on_CancelButton_clicked));

	CancelFlag=false;
	OFDialog=new OverwriteFileDialog(GladeFilename);
}

ImportSectionalsDVD::~ImportSectionalsDVD() {
	delete ProgressWindow;
	delete OFDialog;
}

int ImportSectionalsDVD::Run()
{
	Gtk::FileChooserDialog FileDialog("Browse to the Sectionals directory", Gtk::FILE_CHOOSER_ACTION_SELECT_FOLDER );
	FileDialog.set_current_folder(Cfg->GetSectionalsImportDir());

	/* Add response buttons */
	FileDialog.add_button(Gtk::Stock::CANCEL, Gtk::RESPONSE_CANCEL);
	FileDialog.add_button(Gtk::Stock::OPEN, Gtk::RESPONSE_OK);

	if (FileDialog.run()!=Gtk::RESPONSE_OK)
		return false;
	FileDialog.hide();

	/* Get the Volume Information */
	ImportSectionalsDVDVolumeChooserDialog *VolumeDialog =
		new ImportSectionalsDVDVolumeChooserDialog(GladeFilename,FileDialog.get_current_folder());
	try {
		if (VolumeDialog->Run()!=Gtk::RESPONSE_OK) {
			delete VolumeDialog;
			return false;
		}
	}
	catch (Exception Ex) {
		Ex.Show();
		delete VolumeDialog;
		return false;
	}

	std::string Volume;
	time_t EffFrom,EffTo;
	VolumeDialog->GetData(&Volume,&EffFrom,&EffTo);
	delete VolumeDialog;
	VolumeDialog=NULL;

	VolData=Data->GetVolume(&Volume,&EffFrom,&EffTo);
	if (!VolData) {
		throw Exception("Error creating store for volume data",Exception::LEVEL_ERROR);
	}

	std::string ChartDstDir=VolData->GetDataDir()+"/geocharts/";
	Util::ShellExpand(&ChartDstDir);

	/* Create the dest dir
	 * Ignore errors since it's not really an error for the directory to already exist
	 * and anything else will get caught with the file copy
	 */
	mkdir(ChartDstDir.c_str(),0775);

	SubProgressBar->show();
	SubProgressBar->set_text("Sectionals");
	ProgressBar->set_text("Total");
	ProgressWindow->show();
	ProgressDone=0;

	ImportSectionals(FileDialog.get_current_folder(),ChartDstDir);

	if (!CancelFlag) {
		SubProgressBar->set_text("TACs");
		ImportTACs(FileDialog.get_current_folder()+"/Terminal-Area-Charts",ChartDstDir);
	}
	if (!CancelFlag) {
		SubProgressBar->set_text("Flyways");
		ImportFlyways(FileDialog.get_current_folder()+"/Flyways-not-georeferenced",ChartDstDir);
	}

	if (CancelFlag) {
		return false;
	} else {
		try {
			VolData->Save();
		}
		catch (Exception Ex) {
			Ex.Show();
			return false;
		}
	}

	return true;
}

void ImportSectionalsDVD::ImportSectionals(std::string SrcDir, std::string DstDir)
{
	std::string Search;
	glob_t globdata;
	double NB,EB,SB,WB;
	bool Duplicate;

	/* Find files in this directory */
	Search=SrcDir+"/*.tif";

	if (glob(Search.c_str(),0,NULL,&globdata)==0)
	{
		for (unsigned int i=0;i<globdata.gl_pathc && !CancelFlag;i++)
		{
			SubProgressBar->set_fraction((double)i/globdata.gl_pathc);
			ProgressBar->set_fraction(SectionalsProgressFrac*i/globdata.gl_pathc);

			std::string FileName=globdata.gl_pathv[i];
			FileName=FileName.substr(FileName.find_last_of('/')+1);

			/* Search for duplicates */
			Duplicate=false;
			for (std::vector<GeoChart*>::iterator GCi=VolData->GetGeoCharts()->begin();GCi!=VolData->GetGeoCharts()->end();++GCi) {
				if (FileName.compare((*GCi)->GetFileName().c_str())==0) {
					Duplicate=true;
					break;
				}
			}

			if (!Duplicate) {
				Util::CopyFile(globdata.gl_pathv[i],DstDir,"",OFDialog);
				GeoChart* GC = new GeoChart();

				GC->SetFileName(FileName);
				/* The name is set automatically from the filename */

				SectionalChart* Sectional = new SectionalChart(Cfg,DstDir+FileName);
				if (!Sectional) {
					throw Exception(std::string("Error reading ")+std::string(globdata.gl_pathv[i]),Exception::LEVEL_ERROR);
				}
				Sectional->GetBounds(
						&EB,
						&WB,
						&NB,
						&SB
						);
				GC->SetBounding(NB,EB,SB,WB);
				GC->SetGeoReferenced(true);
				GC->SetType(GeoChart::TYPE_SECTIONAL);
				delete Sectional;

				VolData->GetGeoCharts()->push_back(GC);
			}
		}
		globfree(&globdata);
	}
	ProgressDone+=SectionalsProgressFrac;
}

void ImportSectionalsDVD::ImportTACs(std::string SrcDir, std::string DstDir)
{
	std::string Search;
	glob_t globdata;
	double NB,EB,SB,WB;
	bool Duplicate;

	/* Find files in this directory */
	Search=SrcDir+"/*.tif";

	if (glob(Search.c_str(),0,NULL,&globdata)==0)
	{
		for (unsigned int i=0;i<globdata.gl_pathc && !CancelFlag;i++)
		{
			SubProgressBar->set_fraction((double)i/globdata.gl_pathc);
			ProgressBar->set_fraction(ProgressDone+TACsProgressFrac*i/globdata.gl_pathc);

			std::string FileName=globdata.gl_pathv[i];
			FileName=FileName.substr(FileName.find_last_of('/')+1);

			/* Search for duplicates */
			Duplicate=false;
			for (std::vector<GeoChart*>::iterator GCi=VolData->GetGeoCharts()->begin();GCi!=VolData->GetGeoCharts()->end();++GCi) {
				if (FileName.compare((*GCi)->GetFileName().c_str())==0) {
					Duplicate=true;
					break;
				}
			}

			if (!Duplicate) {
				Util::CopyFile(globdata.gl_pathv[i],DstDir,"",OFDialog);
				GeoChart* GC = new GeoChart();

				GC->SetFileName(FileName);
				/* The name is set automatically from the filename */

				SectionalChart* Sectional = new SectionalChart(Cfg,DstDir+FileName);
				if (!Sectional) {
					throw Exception(std::string("Error reading ")+std::string(globdata.gl_pathv[i]),Exception::LEVEL_ERROR);
				}

				Sectional->GetBounds(
						&EB,
						&WB,
						&NB,
						&SB
						);
				GC->SetBounding(NB,EB,SB,WB);
				GC->SetGeoReferenced(true);
				GC->SetType(GeoChart::TYPE_TAC);
				delete Sectional;

				VolData->GetGeoCharts()->push_back(GC);
			}
		}
		globfree(&globdata);
	}
	ProgressDone+=TACsProgressFrac;
}

void ImportSectionalsDVD::ImportFlyways(std::string SrcDir, std::string DstDir)
{
	std::string Search;
	glob_t globdata;
	bool Duplicate;

	/* Find files in this directory */
	Search=SrcDir+"/*.tif";

	if (glob(Search.c_str(),0,NULL,&globdata)==0)
	{
		for (unsigned int i=0;i<globdata.gl_pathc && !CancelFlag;i++)
		{
			SubProgressBar->set_fraction((double)i/globdata.gl_pathc);
			ProgressBar->set_fraction(ProgressDone+FlywaysProgressFrac*i/globdata.gl_pathc);

			std::string FileName=globdata.gl_pathv[i];
			FileName=FileName.substr(FileName.find_last_of('/')+1);

			/* Search for duplicates */
			Duplicate=false;
			for (std::vector<GeoChart*>::iterator GCi=VolData->GetGeoCharts()->begin();GCi!=VolData->GetGeoCharts()->end();++GCi) {
				if (FileName.compare((*GCi)->GetFileName().c_str())==0) {
					Duplicate=true;
					break;
				}
			}

			if (!Duplicate) {
				Util::CopyFile(globdata.gl_pathv[i],DstDir,"",OFDialog);
				GeoChart* GC = new GeoChart();

				GC->SetFileName(FileName);
				/* The name is set automatically from the filename */

				GC->SetGeoReferenced(false);
				GC->SetType(GeoChart::TYPE_FLYWAY);
				VolData->GetGeoCharts()->push_back(GC);
			}
		}
		globfree(&globdata);
	}
	ProgressDone+=FlywaysProgressFrac;
}

void ImportSectionalsDVD::on_CancelButton_clicked() {
	CancelFlag=true;
}
