#include <unistd.h>
#include <string.h>
#include <glob.h>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <sstream>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <libglademm.h>
#include <gtkmm/messagedialog.h>
#include <gtkmm/stock.h>
#include <gtkmm/main.h>
#include <gtkmm/filechooserdialog.h>
#include "ImportdTPPDVD.h"
#include "Airport.h"
#include "Exception.h"
#include "Util.h"

ImportdTPPDVD::ImportdTPPDVD(Config *Cfg, std::string *GladeFilename, DataStore *Data)
{
	this->Cfg=Cfg;
	this->GladeFilename=GladeFilename;
	this->Data=Data;

	try
	{
		std::string gladefile;

		refXml = Gnome::Glade::Xml::create(*GladeFilename, "ProgressWindow");
	}
	catch (const Gnome::Glade::XmlError& ex)
	{
		Gtk::MessageDialog *message = new Gtk::MessageDialog("Error reading data file.  Program cannot continue.",
			false,Gtk::MESSAGE_ERROR,Gtk::BUTTONS_CANCEL,true);
		message->run();
		Gtk::Main::quit();
	}

	refXml->get_widget("ProgressWindow",ProgressWindow);
	refXml->get_widget("SubProgressBar",SubProgressBar);
	refXml->get_widget("ProgressBar",ProgressBar);
	refXml->get_widget("CancelButton",CancelButton);

	ProgressWindow->set_deletable(false);
	CancelButton->signal_clicked().connect(sigc::mem_fun(this,&ImportdTPPDVD::on_CancelButton_clicked));

	CancelFlag=false;
	BytesRead=0;
	BytesAvail=0;

	OFDialog=new OverwriteFileDialog(GladeFilename);
}

ImportdTPPDVD::~ImportdTPPDVD() {
	delete ProgressWindow;
	delete OFDialog;
}

int ImportdTPPDVD::Run()
{
	int Count;
	struct stat statdata;
	const xmlpp::Node* Root;
	glob_t globdata;
	std::string ImgSrcDir,ImgDstDir;
	double ProgressDone=0.0;

	Gtk::FileChooserDialog FileDialog("Browse to the d-TPP data", Gtk::FILE_CHOOSER_ACTION_OPEN);
	FileDialog.set_current_folder(Cfg->GetdTPPImportDir());

	/* Add filters */
	Gtk::FileFilter FileFilter;
	FileFilter.set_name("Metadata (d-TPP_Metafile.xml)");
	FileFilter.add_pattern("d-TPP_Metafile.xml");
	FileDialog.add_filter(FileFilter);

	/* Add response buttons */
	FileDialog.add_button(Gtk::Stock::CANCEL, Gtk::RESPONSE_CANCEL);
	FileDialog.add_button(Gtk::Stock::OPEN, Gtk::RESPONSE_OK);

	if (FileDialog.run()!=Gtk::RESPONSE_OK)
		return false;

	if (stat(FileDialog.get_filename().c_str(),&statdata)!=0) {
		Gtk::MessageDialog *message = new Gtk::MessageDialog(
			"Error checking d-TPP file.",false,Gtk::MESSAGE_ERROR,
			Gtk::BUTTONS_CANCEL,true);
		message->run();
		delete message;
		return false;
	}
	BytesAvail=statdata.st_size;

	FileDialog.hide();

	/* Load metadata */
	Root=ParseXmlInit(FileDialog.get_filename().c_str());
	if (!Root) {
		Gtk::MessageDialog *message = new Gtk::MessageDialog(
			"Error reading d-TPP file.",false,Gtk::MESSAGE_ERROR,
			Gtk::BUTTONS_CANCEL,true);
		message->run();
		delete message;
		return false;
	}

	ParseXmlMetadata(Root);
	if (!ConfirmData(ImportVolume,ImportEffectiveFrom,ImportEffectiveTo))
		return false;

	ImportEffectiveTo--;	/* We subtract 1 second to avoid any overlap */
	VolData=Data->GetVolume(&ImportVolume,&ImportEffectiveFrom,&ImportEffectiveTo);
	if (!VolData) {
		throw Exception("Error creating store for volume data",Exception::LEVEL_ERROR);
	}

	if (VolData->GetAirports()->size()==0) {
		Gtk::MessageDialog *message = new Gtk::MessageDialog(
			"No airports exist in this volume - import the flight database first.",false,Gtk::MESSAGE_ERROR,
			Gtk::BUTTONS_CANCEL,true);
		message->run();
		delete message;
		return false;
	}

	/* Clear the existing datastore */
	VolData->ClearApproachPlates();

	SubProgressBar->show();
	SubProgressBar->set_text("Parsing");
	ProgressBar->set_text("Total");
	ProgressWindow->show();

	/* Copy the image files */
	ImgSrcDir=FileDialog.get_filename();
	ImgSrcDir.erase(ImgSrcDir.find_last_of('/'));	/* Erase the xml file name */
	ImgSrcDir.erase(ImgSrcDir.find_last_of('/'));	/* Erase the xml subdirectory text */
	ImgSrcDir+=std::string("/published_pdfs/");	/* Add the source image directory */

	ImgDstDir=VolData->GetDataDir()+"/plateimages/";
	Util::ShellExpand(&ImgDstDir);

	/* Create the dest dir
	 * Ignore errors since it's not really an error for the directory to already exist
	 * and anything else will get caught with the file copy
	 */
	mkdir(ImgDstDir.c_str(),0775);

	/* Actually do the copy */

	/* Parse XML source data */
	Count=0;
	xmlpp::Node::NodeList States = Root->get_children();
	for(
		xmlpp::Node::NodeList::iterator Statei = States.begin();
		Statei != States.end() && !CancelFlag;
		Statei++) {
		ParseXmlState(*Statei);
		Count++;
		SubProgressBar->set_fraction((double)Count/States.size());
		ProgressBar->set_fraction(ProgressDone+ParseProgressFrac*Count/States.size());
		Util::ScheduleMainLoop();
	}

	/* The glob() can take a while, so make sure the ProgressDialog is shown */
	Util::ScheduleMainLoop();

	SubProgressBar->set_text("Copying");
	if (glob((ImgSrcDir+Util::GlobiPattern("*.PDF")).c_str(),GLOB_NOSORT,NULL,&globdata)==0) {

		/* Copy all other files that are associated with any loaded approach at any loaded airport */
		unsigned int i;
		for (i=0;i<globdata.gl_pathc && !CancelFlag;i++) {
			std::string Filename;	/* Just the file name with no path */
			Filename=globdata.gl_pathv[i];
			if (Filename.find_last_of('/')!=Filename.npos) {
				Filename=Filename.substr(Filename.find_last_of('/')+1);
			}
			std::string UpperFilename;	/* Converted to upper case */
			UpperFilename=Filename;
			std::transform(UpperFilename.begin(), UpperFilename.end(), UpperFilename.begin(), toupper);
			if (VolData->IsPlateImageFileUsed(UpperFilename) || UpperFilename==FRONTMATTERFILENAME) {
				Util::CopyFile(globdata.gl_pathv[i],ImgDstDir,UpperFilename,OFDialog);
				SubProgressBar->set_fraction((double)i/globdata.gl_pathc);
				ProgressBar->set_fraction(CopyProgressFrac*i/globdata.gl_pathc);
			}
		}
		globfree(&globdata);
	} else {
		VolData->ClearApproachPlates();
		throw Exception("Error reading approach plate filenames",Exception::LEVEL_ERROR);
	}
	ProgressDone=CopyProgressFrac;

	/* We have now loaded all the ApproachPlates */

	if (CancelFlag) {
		VolData->ClearApproachPlates();
		return false;
	} else {
		try {
			VolData->Save();
		}
		catch (Exception Ex) {
			VolData->ClearApproachPlates();
			Ex.Show();
			return false;
		}
	}

	return true;
}

void ImportdTPPDVD::on_CancelButton_clicked() {
	CancelFlag=true;
}

bool ImportdTPPDVD::ConfirmData(std::string Volume, time_t EffectiveFrom, time_t EffectiveTo) {
	struct tm bdtime;
	std::string CycleStr;
	std::ostringstream EffFromOStr,EffToOStr;
	bool result=false;

	CycleStr="Volume="+Volume+"\n";
	gmtime_r(&EffectiveFrom,&bdtime);
	EffFromOStr<<"Eff From="<<std::setfill('0')<<
		std::setw(2)<<bdtime.tm_hour<<std::setw(2)<<bdtime.tm_min<<"Z "<<
		std::setw(2)<<(bdtime.tm_mon+1)<<"/"<<
		std::setw(2)<<bdtime.tm_mday<<"/"<<
		std::setw(2)<<(bdtime.tm_year-100)<<"\n";
	gmtime_r(&EffectiveTo,&bdtime);
	EffToOStr<<"Eff To="<<std::setfill('0')<<
		std::setw(2)<<bdtime.tm_hour<<std::setw(2)<<bdtime.tm_min<<"Z "<<
		std::setw(2)<<(bdtime.tm_mon+1)<<"/"<<
		std::setw(2)<<bdtime.tm_mday<<"/"<<
		std::setw(2)<<(bdtime.tm_year-100)<<"\n";

	Gtk::MessageDialog *message = new Gtk::MessageDialog(
		CycleStr+EffFromOStr.str()+EffToOStr.str(),false,Gtk::MESSAGE_QUESTION,
		Gtk::BUTTONS_OK_CANCEL,true);
	message->set_title("Confirm");

	result=(message->run()==Gtk::RESPONSE_OK);
	delete message;

	return result;
}

const xmlpp::Node* ImportdTPPDVD::ParseXmlInit(std::string Filename) {
	try {
		Parser.set_substitute_entities(); //We just want the text to be resolved/unescaped automatically.
		Parser.parse_file(Filename);
	} catch (const std::exception& ex) {
		fprintf(stderr,"Exception caught: %s\n",ex.what());
		return NULL;
	}

	const xmlpp::Node* TPP = Parser.get_document()->get_root_node();

	if (TPP->get_name()!="digital_tpp")
		return NULL;

	return TPP;
}

void ImportdTPPDVD::ParseXmlMetadata(const xmlpp::Node* Root) {
	if(const xmlpp::Element* TPPElement = dynamic_cast<const xmlpp::Element*>(Root)) {
		const xmlpp::Element::AttributeList& TPPAttributes = TPPElement->get_attributes();
		for(xmlpp::Element::AttributeList::const_iterator TPPElementi = TPPAttributes.begin(); TPPElementi != TPPAttributes.end(); ++TPPElementi) {
			const xmlpp::Attribute* attribute = *TPPElementi;
			if (attribute->get_name()=="cycle") {
				ImportVolume=attribute->get_value();
			} else if (attribute->get_name()=="from_edate") {
				struct tm bdtime;
				memset(&bdtime,0,sizeof(bdtime));

				if (sscanf(attribute->get_value().c_str(),"%02u%02uZ %02u/%02u/%02u",
					&(bdtime.tm_hour),
					&(bdtime.tm_min),
					&(bdtime.tm_mon),
					&(bdtime.tm_mday),
					&(bdtime.tm_year))==5) {
					bdtime.tm_mon--;
					bdtime.tm_year+=100;
					ImportEffectiveFrom=Util::GMMkTime(&bdtime);
				}
			} else if (attribute->get_name()=="to_edate") {
				struct tm bdtime;
				memset(&bdtime,0,sizeof(bdtime));
				if (sscanf(attribute->get_value().c_str(),"%02u%02uZ %02u/%02u/%02u",
					&(bdtime.tm_hour),
					&(bdtime.tm_min),
					&(bdtime.tm_mon),
					&(bdtime.tm_mday),
					&(bdtime.tm_year))==5) {
					bdtime.tm_mon--;
					bdtime.tm_year+=100;
					ImportEffectiveTo=Util::GMMkTime(&bdtime);
				}
			}
		}
	}
}

void ImportdTPPDVD::ParseXmlState(const xmlpp::Node* State) {
	if (State->get_name()!="state_code")
		return;

	if(const xmlpp::Element* StateElement = dynamic_cast<const xmlpp::Element*>(State)) {
		const xmlpp::Element::AttributeList& attributes = StateElement->get_attributes();

		for(xmlpp::Element::AttributeList::const_iterator StateElementi = attributes.begin(); StateElementi != attributes.end(); ++StateElementi) {
			const xmlpp::Attribute* attribute = *StateElementi;

			if (attribute->get_name()=="ID")
				Partial.SetStateID(attribute->get_value());
			else if (attribute->get_name()=="state_fullname")
				Partial.SetStateName(attribute->get_value());
		}

		xmlpp::Node::NodeList Cities = State->get_children();
		for(xmlpp::Node::NodeList::iterator Cityi = Cities.begin(); Cityi != Cities.end(); ++Cityi) {
			ParseXmlCity(*Cityi);
		}
	}
}

void ImportdTPPDVD::ParseXmlCity(const xmlpp::Node* City) {
	if (City->get_name()!="city_name")
		return;

	if(const xmlpp::Element* CityElement = dynamic_cast<const xmlpp::Element*>(City)) {
		const xmlpp::Element::AttributeList& attributes = CityElement->get_attributes();

		for(xmlpp::Element::AttributeList::const_iterator CityElementi = attributes.begin(); CityElementi != attributes.end(); ++CityElementi) {
			const xmlpp::Attribute* attribute = *CityElementi;

			if (attribute->get_name()=="ID")
				Partial.SetCityName(attribute->get_value());
			else if (attribute->get_name()=="volume")
				Partial.SetVolume(attribute->get_value());
		}

		xmlpp::Node::NodeList Airports = City->get_children();
		for(xmlpp::Node::NodeList::iterator Airporti = Airports.begin(); Airporti != Airports.end(); ++Airporti) {
			ParseXmlAirport(*Airporti);
		}
	}
}

void ImportdTPPDVD::ParseXmlAirport(const xmlpp::Node* Airport) {
	if (Airport->get_name()!="airport_name")
		return;

	if(const xmlpp::Element* AirportElement = dynamic_cast<const xmlpp::Element*>(Airport)) {
		const xmlpp::Element::AttributeList& attributes = AirportElement->get_attributes();

		for(xmlpp::Element::AttributeList::const_iterator AirportElementi = attributes.begin(); AirportElementi != attributes.end(); ++AirportElementi) {
			const xmlpp::Attribute* attribute = *AirportElementi;

			if (attribute->get_name()=="ID")
				Partial.SetAirportID(attribute->get_value());
			else if (attribute->get_name()=="military")
				Partial.SetAirportMilitary(attribute->get_value());
			else if (attribute->get_name()=="apt_ident")
				Partial.SetAirportIdent(attribute->get_value());
			else if (attribute->get_name()=="icao_ident")
				Partial.SetAirportICAOIdent(attribute->get_value());
		}

		::Airport *DSAirport=NULL;
		/* Find matching Airport by ICAO identifier */
		VolData->GetAirportByICAO(Partial.GetAirportICAOIdent());
		if (!DSAirport) {
			/* Find matching Airport by IATA identifier */
			DSAirport=VolData->GetAirportByIATA(Partial.GetAirportIdent());
		}
		if (!DSAirport) {
			/* A few airports have the IATA identifier in the ICAO field - look there last */
			DSAirport=VolData->GetAirportByICAO(Partial.GetAirportIdent());
		}
		if (!DSAirport) {
			/* TODO: Implement heliport support so we don't have to ignore their approaches */
			/* Silently ignore approaches to unknown airports (probably heliports) */
		} else {
			for (
				std::vector<ApproachPlate*>::iterator Platei=DSAirport->GetApproachPlates()->begin();
				Platei!=DSAirport->GetApproachPlates()->end();
				++Platei)
				delete (*Platei);
			DSAirport->GetApproachPlates()->clear();
			xmlpp::Node::NodeList Records = Airport->get_children();
			for(xmlpp::Node::NodeList::iterator Recordi = Records.begin(); Recordi != Records.end(); ++Recordi) {
				ParseXmlRecord(*Recordi,DSAirport);
			}
		}
	}
}

void ImportdTPPDVD::ParseXmlRecord(const xmlpp::Node* Record, Airport *DSAirport) {
	if (Record->get_name()!="record")
		return;

	xmlpp::Node::NodeList Tags = Record->get_children();

	for(xmlpp::Node::NodeList::iterator Tagsi = Tags.begin(); Tagsi != Tags.end(); ++Tagsi) {
		if (const xmlpp::Element* TagElement = dynamic_cast<const xmlpp::Element*>(*Tagsi)) {
			if (TagElement->has_child_text()) {
				std::string Content=TagElement->get_child_text()->get_content();

				if ((*Tagsi)->get_name()=="chartseq") {
					Partial.SetSeq(Content);
				} else if ((*Tagsi)->get_name()=="chart_code") {
					Partial.SetCode(Content);
				} else if ((*Tagsi)->get_name()=="chart_name") {
					Partial.SetName(Content);
				} else if ((*Tagsi)->get_name()=="useraction") {
					Partial.SetUserAction(Content);
				} else if ((*Tagsi)->get_name()=="pdf_name") {
					std::transform(Content.begin(), Content.end(), Content.begin(), toupper);
					Partial.SetImgName(Content);
					Partial.SetImgType(ApproachPlate::PDF);
				} else if ((*Tagsi)->get_name()=="cn_flg") {
					Partial.SetCNFlag(Content);
				} else if ((*Tagsi)->get_name()=="cnsection") {
					Partial.SetCNSection(Content);
				} else if ((*Tagsi)->get_name()=="cnpage") {
					Partial.SetCNPage(Content);
				} else if ((*Tagsi)->get_name()=="bvsection") {
					Partial.SetBVSection(Content);
				} else if ((*Tagsi)->get_name()=="bvpage") {
					Partial.SetBVPage(Content);
				} else if ((*Tagsi)->get_name()=="procuid") {
					Partial.SetProcUID(Content);
				} else if ((*Tagsi)->get_name()=="two_colored") {
					Partial.SetTwoColored(Content);
				} else if ((*Tagsi)->get_name()=="civil") {
					Partial.SetCivil(Content);
				} else if ((*Tagsi)->get_name()=="faanfd15") {
					Partial.SetFAANFD15(Content);
				} else if ((*Tagsi)->get_name()=="faanfd18") {
					Partial.SetFAANFD18(Content);
				} else if ((*Tagsi)->get_name()=="copter") {
					Partial.SetCopter(Content);
				}
			}
		}
	}
	ApproachPlate *NewPlate = new ApproachPlate();
	NewPlate->Copy(&Partial);

	DSAirport->AddApproachPlate(NewPlate);
}
