#include <sstream>
#include <ostream>
#include <iomanip>
#include <iostream>
#include <fstream>
#include <algorithm>
#include <cstring>
#include <ctime>
#include <cstdlib>
#include "../config.h"
#include "METARs.h"
#include "METAR.h"
#include "Util.h"

/* Information on how to categorize and draw METAR information was taken from
 * the following pages on Nov 17 2008:
 * http://adds.aviationweather.gov/metars/description2.php
 * http://adds.aviationweather.gov/metars/description_ifr.php
 */

METARs::METARs(Config *Cfg)
{
	CurlData=NULL;
	CurlDataSize=0;
	this->Cfg=Cfg;
}

METARs::~METARs()
{
}

/* Offset is in hours from current time */
std::string METARs::CycleURL(short Offset) {
	std::ostringstream ostr;
	struct tm *bdtime;
	time_t currtime;

	currtime=time(NULL);
	bdtime=gmtime(&currtime);

	ostr<<Cfg->GetMETARSourceURL()<<
		std::setw(2)<<std::setfill('0')<<(bdtime->tm_hour+Offset)<<
		"Z.TXT";

	return ostr.str();
}

/* Get the data from the last cycle (which is complete but possibly stale)
 * plus the current cycle (which is fresh but possibly incomplete) to ensure
 * we have the latest available, most complete data.
 */
bool METARs::Download() {
	Stations.clear();
	if (!DownloadCycleFileURL(CycleURL(0)))
		return false;
	if (!DownloadCycleFileURL(CycleURL(-1)))
		return false;

	CleanUp();
	Save();

	return true;
}

bool METARs::DownloadCycleFileURL(std::string URL) {
	CURL *Curl=curl_easy_init();
	CURLcode Result;

	if (!Curl)
		return false;

	Result=curl_easy_setopt(Curl,CURLOPT_URL,URL.c_str());
	if (Result!=CURLE_OK) {
		CurlCleanup(Curl);
		return false;
	}

	Result=curl_easy_setopt(Curl,CURLOPT_WRITEFUNCTION,CurlWriteHandler);
	if (Result!=CURLE_OK) {
		CurlCleanup(Curl);
		return false;
	}

	Result=curl_easy_setopt(Curl,CURLOPT_WRITEDATA,this);
	if (Result!=CURLE_OK) {
		CurlCleanup(Curl);
		return false;
	}

	Result=curl_easy_perform(Curl);
	if (Result!=CURLE_OK) {
		CurlCleanup(Curl);
		return false;
	}

	if (CurlDataSize==0) {
		CurlCleanup(Curl);
		return false;
	}

	/* At this point the CurlData buffer is filled with downloaded data */
	size_t LineStart=0, LineEnd;
	std::string Line;

	while (LineStart<CurlDataSize) {
		for (LineEnd=LineStart;LineEnd<CurlDataSize;LineEnd++) {
			if (CurlData[LineEnd]=='\r' || CurlData[LineEnd]=='\n') {
				break;
			}
		}
		/* LineEnd is now an index to either the end of the line or the
		 * end of the data
		 */
		Line.assign(&(CurlData[LineStart]),LineEnd-LineStart);	/* Omit the EOL marker */
		if (Line.size()>5 && Line.substr(4,1).compare("/")!=0) {
			/* Line is proper data - not a blank line or timestamp */

			/* Trim any trailing whitespace */
			size_t endpos = Line.find_last_not_of(" \t\r\n");
			Line=Line.substr(0,endpos+1);

			METAR *WX = new METAR(Line);

			/* Left for debugging METAR parsing as needed
			if (WX->GetValidParseFlag())
				printf("Station=%s Time=%lu Vis=%0.2f Ceiling=%d\n",WX->GetStation().c_str(),WX->GetZuluTime(),WX->GetVisibility(),WX->GetCeiling());
			else
				printf("Station=%s INVALID:%s\n",WX->GetStation().c_str(),WX->GetText().c_str());
			*/

			Stations.push_back(*WX);
		}

		LineStart=LineEnd;
		while ( (CurlData[LineStart]=='\r' || CurlData[LineStart]=='\n') && LineStart<CurlDataSize)
			LineStart++;
	}

	CurlCleanup(Curl);
	return true;
}

void METARs::CurlCleanup(CURL *Curl) {
	if (CurlData)
		free(CurlData);
	CurlData=NULL;
	CurlDataSize=0;

	curl_easy_cleanup(Curl);
}

size_t METARs::CurlWriteHandler(void *ptr, size_t size, size_t nmemb, METARs *stream) {
	void *TmpCurlData;

	/* Make room for the new data */
	TmpCurlData=realloc(stream->CurlData,stream->CurlDataSize+(size*nmemb));
	if (!TmpCurlData)
		return 0;
	stream->CurlData=(char*)TmpCurlData;

	/* Copy the new data into the buffer */
	memcpy(&(stream->CurlData[stream->CurlDataSize]),ptr,size*nmemb);
	stream->CurlDataSize+=size*nmemb;

	return size*nmemb;
}

std::list<METAR> *METARs::GetStations() {
	return &Stations;
}

size_t METARs::GetStationCount() {
	return Stations.size();
}

void METARs::CleanUp() {
	/* Sort the METARs by station, then ZuluTime to optimize later processing */
	Stations.sort();
	Stations.unique();
}

bool METARs::Save() {
	std::ofstream ofs;
	std::string METARFilename;

	METARFilename=Cfg->GetUserDataDir()+std::string("/")+
		std::string("METAR.")+std::string(PACKAGE);
	Util::ShellExpand(&METARFilename);
	ofs.open(METARFilename.c_str(),std::ios::out);

	for (std::list<METAR>::iterator METARi=Stations.begin(); METARi!=Stations.end(); ++METARi) {
		ofs<<METARi->GetText()<<std::endl;
	}
	ofs.close();

	return true;
}

bool METARs::Load() {
	std::ifstream ifs;
	std::string METARFilename,Line;

	METARFilename=Cfg->GetUserDataDir()+std::string("/")+
		std::string("METAR.")+std::string(PACKAGE);
	Util::ShellExpand(&METARFilename);
	ifs.open(METARFilename.c_str(),std::ios::in);

	if (ifs.fail()) {
//TODO Show error
		return false;
	}

	Stations.clear();
	while (!ifs.eof() && !ifs.fail()) {
		getline(ifs,Line);
		if (Line.size()>5 && Line.substr(4,1).compare("/")!=0) {
			/* Line is proper data - not a blank line or timestamp */
			METAR *WX = new METAR(Line);

			Stations.push_back(*WX);
		}
	}
	ifs.close();

	CleanUp();

	return true;
}

void METARs::Clear() {
	Stations.clear();
}
