#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <string>
#include <sstream>
#include <gtkmm/drawingarea.h>
#include <gtkmm/messagedialog.h>
#include <gtkmm/filechooserdialog.h>
#include <gtkmm/stock.h>
#include <gdkmm/rectangle.h>
#include <gtkmm/main.h>

#include "../config.h"
#include "MainWindow.h"
#include "AboutDialog.h"
#include "GotoDialog.h"
#include "DisclaimerDialog.h"
#include "ApproachSearchDialog.h"
#include "ConfigDialog.h"
#include "Exception.h"
#include "Util.h"
#include "ChartChooserDialog.h"
#include "DataManagerDialog.h"
#include "WelcomeDialog.h"
#include "AircraftDialog.h"
#include "FlashlightDialog.h"

MainWindow::MainWindow(App *Application, std::string *GladeFilename, Config *Cfg)
{
	this->Application=Application;
	this->GladeFilename=GladeFilename;
	this->Cfg=Cfg;

	try
	{
		refXml = Gnome::Glade::Xml::create(*GladeFilename, "MainWindow");
	}
	catch (const Gnome::Glade::XmlError& ex)
	{
		Gtk::MessageDialog *message = new Gtk::MessageDialog("Error reading data file.  Program cannot continue.",
			false,Gtk::MESSAGE_ERROR,Gtk::BUTTONS_CANCEL,true);
		message->run();
		delete message;
		Gtk::Main::quit();
	}

	refXml->get_widget("MainWindow",GtkWindow);
	GtkWindow->hide();

	/* Set the default icon */
	try {
		Glib::RefPtr<Gdk::Pixbuf> appicon = Gdk::Pixbuf::create_from_file(std::string(DATADIR)+"/pixmaps/vifiricon.png");
		GtkWindow->set_icon(appicon);
		GtkWindow->set_default_icon(appicon);
	}
	catch (const Glib::FileError& ex) { /* Ignore errors */}

	GtkWindow->show();
	GtkWindow->maximize();

	refXml->get_widget_derived("DrawingArea",DrawingArea);
	refXml->get_widget("FileNewAircraft",FileNewAircraft);
	refXml->get_widget("FileNewRoute",FileNewRoute);
	refXml->get_widget("FileNewWxData",FileNewWxData);
	refXml->get_widget("FileOpenAircraft",FileOpenAircraft);
	refXml->get_widget("FileOpenChart",FileOpenChart);
	refXml->get_widget("FileOpenRoute",FileOpenRoute);
	refXml->get_widget("FileOpenWxData",FileOpenWxData);
	refXml->get_widget("FileClose",FileClose);
	refXml->get_widget("FileDataManager",FileDataManager);
	refXml->get_widget("FileConfigure",FileConfigure);
	refXml->get_widget("FileFlashlight",FileFlashlight);
	refXml->get_widget("FileQuit",FileQuit);

	refXml->get_widget("EditAircraft",EditAircraft);
	refXml->get_widget("EditRoute",EditRoute);

	refXml->get_widget("View",ViewMenu);
	refXml->get_widget("ViewZoomIn",ViewZoomIn);
	refXml->get_widget("ViewZoomOut",ViewZoomOut);
	refXml->get_widget("ViewZoom30Pct",ViewZoom30Pct);
	refXml->get_widget("ViewZoom60Pct",ViewZoom60Pct);
	refXml->get_widget("ViewZoom100Pct",ViewZoom100Pct);
	refXml->get_widget("ViewZoom150Pct",ViewZoom150Pct);
	refXml->get_widget("ViewZoom200Pct",ViewZoom200Pct);
	refXml->get_widget("ViewZoom400Pct",ViewZoom400Pct);
	refXml->get_widget("ViewPanLeft",ViewPanLeft);
	refXml->get_widget("ViewPanRight",ViewPanRight);
	refXml->get_widget("ViewPanUp",ViewPanUp);
	refXml->get_widget("ViewPanDown",ViewPanDown);
	refXml->get_widget("ViewPanFastLeft",ViewPanFastLeft);
	refXml->get_widget("ViewPanFastRight",ViewPanFastRight);
	refXml->get_widget("ViewPanFastUp",ViewPanFastUp);
	refXml->get_widget("ViewPanFastDown",ViewPanFastDown);
	refXml->get_widget("ViewGoto",ViewGoto);
	refXml->get_widget("ViewFullscreen",ViewFullscreen);

	refXml->get_widget("SearchApproaches",SearchApproaches);

	refXml->get_widget("GPSActive",GPSActive);
	refXml->get_widget("GPSStartgpsd",GPSStartgpsd);
	refXml->get_widget("GPSAutoSelectChart",GPSAutoSelectChart);

	refXml->get_widget("HelpAbout",HelpAbout);
	refXml->get_widget("HelpDisclaimer",HelpDisclaimer);

	refXml->get_widget("Statusbar",Statusbar);

	FileNewAircraft->signal_activate().connect(sigc::mem_fun(this,&MainWindow::on_FileNewAircraft_activate));
	FileNewRoute->signal_activate().connect(sigc::mem_fun(this,&MainWindow::on_FileNewRoute_activate));
	FileNewWxData->signal_activate().connect(sigc::mem_fun(this,&MainWindow::on_FileNewWxData_activate));
	FileOpenAircraft->signal_activate().connect(sigc::mem_fun(this,&MainWindow::on_FileOpenAircraft_activate));
	FileOpenChart->signal_activate().connect(sigc::mem_fun(this,&MainWindow::on_FileOpenChart_activate));
	FileOpenRoute->signal_activate().connect(sigc::mem_fun(this,&MainWindow::on_FileOpenRoute_activate));
	FileOpenWxData->signal_activate().connect(sigc::mem_fun(this,&MainWindow::on_FileOpenWxData_activate));
	FileClose->signal_activate().connect(sigc::mem_fun(this,&MainWindow::on_FileClose_activate));
	FileDataManager->signal_activate().connect(sigc::mem_fun(this,&MainWindow::on_FileDataManager_activate));
	FileConfigure->signal_activate().connect(sigc::mem_fun(this,&MainWindow::on_FileConfigure_activate));
	FileFlashlight->signal_activate().connect(sigc::mem_fun(this,&MainWindow::on_FileFlashlight_activate));
	FileQuit->signal_activate().connect(sigc::mem_fun(this,&MainWindow::on_FileQuit_activate));

	EditAircraft->signal_activate().connect(sigc::mem_fun(this,&MainWindow::on_EditAircraft_activate));
	EditRoute->signal_activate().connect(sigc::mem_fun(this,&MainWindow::on_EditRoute_activate));

	ViewZoomIn->signal_activate().connect(sigc::mem_fun(this,&MainWindow::on_ViewZoomIn_activate));
	ViewZoomOut->signal_activate().connect(sigc::mem_fun(this,&MainWindow::on_ViewZoomOut_activate));
	ViewZoom30Pct->signal_activate().connect(sigc::mem_fun(this,&MainWindow::on_ViewZoom30Pct_activate));
	ViewZoom60Pct->signal_activate().connect(sigc::mem_fun(this,&MainWindow::on_ViewZoom60Pct_activate));
	ViewZoom100Pct->signal_activate().connect(sigc::mem_fun(this,&MainWindow::on_ViewZoom100Pct_activate));
	ViewZoom150Pct->signal_activate().connect(sigc::mem_fun(this,&MainWindow::on_ViewZoom150Pct_activate));
	ViewZoom200Pct->signal_activate().connect(sigc::mem_fun(this,&MainWindow::on_ViewZoom200Pct_activate));
	ViewZoom400Pct->signal_activate().connect(sigc::mem_fun(this,&MainWindow::on_ViewZoom400Pct_activate));
	ViewPanLeft->signal_activate().connect(sigc::mem_fun(this,&MainWindow::on_ViewPanLeft_activate));
	ViewPanRight->signal_activate().connect(sigc::mem_fun(this,&MainWindow::on_ViewPanRight_activate));
	ViewPanUp->signal_activate().connect(sigc::mem_fun(this,&MainWindow::on_ViewPanUp_activate));
	ViewPanDown->signal_activate().connect(sigc::mem_fun(this,&MainWindow::on_ViewPanDown_activate));
	ViewPanFastLeft->signal_activate().connect(sigc::mem_fun(this,&MainWindow::on_ViewPanFastLeft_activate));
	ViewPanFastRight->signal_activate().connect(sigc::mem_fun(this,&MainWindow::on_ViewPanFastRight_activate));
	ViewPanFastUp->signal_activate().connect(sigc::mem_fun(this,&MainWindow::on_ViewPanFastUp_activate));
	ViewPanFastDown->signal_activate().connect(sigc::mem_fun(this,&MainWindow::on_ViewPanFastDown_activate));
	ViewGoto->signal_activate().connect(sigc::mem_fun(this,&MainWindow::on_ViewGoto_activate));
	ViewFullscreen->signal_activate().connect(sigc::mem_fun(this,&MainWindow::on_ViewFullscreen_activate));

	/* Allow use of the standard +/- keys for zoom in/out in addition to the keypad keys from the glade file */
	ViewZoomIn->add_accelerator("activate", GtkWindow->get_accel_group(), '+', (Gdk::ModifierType)0, Gtk::ACCEL_VISIBLE);
	ViewZoomOut->add_accelerator("activate", GtkWindow->get_accel_group(), '-', (Gdk::ModifierType)0, Gtk::ACCEL_VISIBLE);

	SearchApproaches->signal_activate().connect(sigc::mem_fun(this,&MainWindow::on_SearchApproaches_activate));

	GPSActive->signal_activate().connect(sigc::mem_fun(this,&MainWindow::on_GPSActive_activate));
	GPSStartgpsd->signal_activate().connect(sigc::mem_fun(this,&MainWindow::on_GPSStartgpsd_activate));

	HelpAbout->signal_activate().connect(sigc::mem_fun(this,&MainWindow::on_HelpAbout_activate));
	HelpDisclaimer->signal_activate().connect(sigc::mem_fun(this,&MainWindow::on_HelpDisclaimer_activate));

	GPSAutoSelectChart->set_active(Cfg->GetAutoSelectChart());

	DrawingArea->SetConfig(Cfg);
	DrawingArea->SetRoute(&RouteData);

	Statusbar->pop();
	Statusbar->push("Vifir");

	FileClose->set_sensitive(false);
	EditAircraft->set_sensitive(false);
	EditRoute->set_sensitive(false);
	ViewMenu->set_sensitive(false);

	DrawingArea->SetStatusbar(Statusbar);

	LastSectionalChange=0;

	Gps=new GPS(Cfg);

	DBusInhibitSSProxy=NULL;
	DBusInhibitPMProxy=NULL;
	DBusInhibitSSCookie=NULL;
	DBusInhibitPMCookie=NULL;

	DBusInit();

	try {
		Util::ChkDir(Cfg->GetDataStoreDir());
	}
	catch (Exception Ex) {
		Ex.Show();
	}

	try {
		Data=new DataStore(GladeFilename,Cfg);

		Data->ScanVolumes();
		Data->ActivateCurrentVolume();
	}
	catch (Exception Ex) {
		Ex.Show();
	}

	/* The Offsets are set to 0,0 when a file is loaded since there's
	 * usually no preferred location.  Override this on startup since
	 * the shutdown location is the preferred location.
	 */
	int XOffset,YOffset;
	XOffset=Cfg->GetChartXOffset();
	YOffset=Cfg->GetChartYOffset();
	if (Cfg->GetOpenChartFile().length()) {
		try {
			OpenGeoChart(Cfg->GetOpenChartFile());
		} catch (Exception Ex) {
			Ex.Show();
		}
	}
	if (Cfg->GetOpenRouteFile().length()) {
		try {
			RouteData.Load(Cfg->GetOpenRouteFile());
			EditRoute->set_sensitive();
		} catch (Exception Ex) {
			Ex.Show();
		}
	}
	if (Cfg->GetOpenAircraftFile().length()) {
		try {
			AircraftData=new Aircraft();
			AircraftData->Load(Cfg->GetOpenAircraftFile());
			EditAircraft->set_sensitive();
		} catch (Exception Ex) {
			delete AircraftData;
			AircraftData=NULL;
			Cfg->SetOpenAircraftFile("");
			Ex.Show();
		}
	} else {
		AircraftData=NULL;
	}

	DrawingArea->SetDisplayScale(Cfg->GetChartScale());
	Cfg->SetChartXOffset(XOffset);
	Cfg->SetChartYOffset(YOffset);
	DrawingArea->SetOffset(Cfg->GetChartXOffset(),Cfg->GetChartYOffset());

	if (Cfg->GetShowWelcomeOnStartup()) {
		WelcomeDialog *Dialog = new WelcomeDialog(GladeFilename,Cfg);
		Dialog->Run();
		delete Dialog;
	}

	ShapefileData=new Shapefile();
	try {
		ShapefileData->Load(Cfg->GetShapeFile());
	} catch (Exception Ex) {
		Ex.Show();
	}
	DrawingArea->SetShapefile(ShapefileData);

#ifndef ENABLEBLEEDINGEDGE
	FileNewWxData->hide();
	FileOpenWxData->hide();
#else
#endif

	CancelFlag=false;
}

MainWindow::~MainWindow() {
	if (Gps)
		delete Gps;
	if (Data)
		delete Data;
	if (AircraftData)
		delete AircraftData;
	if (ShapefileData)
		delete ShapefileData;
	if (Cfg->GetSaveConfigOnExit())
		Cfg->Save();

	DBusShutdown();
}

void MainWindow::on_FileNewAircraft_activate() {
	std::string Filename;
	struct stat statdata;

	std::string DataDir;
	Gtk::FileChooserDialog FileDialog("New Aircraft Filename", Gtk::FILE_CHOOSER_ACTION_SAVE);
	DataDir=Cfg->GetUserDataDir();
	Util::ShellExpand(&DataDir);
	FileDialog.set_current_folder(DataDir);

	/* Add filters */
	Gtk::FileFilter FileFilter;
	FileFilter.set_name("Aircraft (*.aircraft)");
	FileFilter.add_pattern("*.aircraft");
	FileDialog.add_filter(FileFilter);

	/* Add response buttons */
	FileDialog.add_button(Gtk::Stock::CANCEL, Gtk::RESPONSE_CANCEL);
	FileDialog.add_button(Gtk::Stock::OPEN, Gtk::RESPONSE_OK);

	if (FileDialog.run()!=Gtk::RESPONSE_OK)
		return;

	FileDialog.hide();

	Filename=FileDialog.get_filename();
	if (Filename.size()<sizeof(".aircraft") ||
		Filename.substr(Filename.size()-sizeof(".aircraft")+1).compare(".aircraft")!=0) {
		Filename.append(".aircraft");	/* Append .route if it's not already there */
	}

	if (stat(Filename.c_str(),&statdata)==0) {
		Gtk::MessageDialog *Message = new Gtk::MessageDialog("File already exists - overwrite?",
			false,Gtk::MESSAGE_WARNING,Gtk::BUTTONS_OK_CANCEL,true);
		if (Message->run()!=Gtk::RESPONSE_OK) {
			delete Message;
			return;
		}
		delete Message;
	}

	if (!AircraftData) {
		AircraftData=new Aircraft();
	}
	AircraftDialog *Dialog=new AircraftDialog(GladeFilename,AircraftData);
	Dialog->Run();
	delete Dialog;

	try {
		AircraftData->Save(Filename);
	}
	catch (Exception Ex) {
		Ex.Show();
	}

	Cfg->SetOpenAircraftFile(Filename);
	EditAircraft->set_sensitive();
	DrawingArea->Redraw();
}

void MainWindow::on_FileNewRoute_activate() {
	std::string Filename;
	struct stat statdata;

	std::string DataDir;
	Gtk::FileChooserDialog FileDialog("New Route Filename", Gtk::FILE_CHOOSER_ACTION_SAVE);
	DataDir=Cfg->GetUserDataDir();
	Util::ShellExpand(&DataDir);
	FileDialog.set_current_folder(DataDir);

	/* Add filters */
	Gtk::FileFilter FileFilter;
	FileFilter.set_name("Route (*.route)");
	FileFilter.add_pattern("*.route");
	FileDialog.add_filter(FileFilter);

	/* Add response buttons */
	FileDialog.add_button(Gtk::Stock::CANCEL, Gtk::RESPONSE_CANCEL);
	FileDialog.add_button(Gtk::Stock::OPEN, Gtk::RESPONSE_OK);

	if (FileDialog.run()!=Gtk::RESPONSE_OK)
		return;

	Filename=FileDialog.get_filename();
	if (Filename.size()<sizeof(".route") ||
		Filename.substr(Filename.size()-sizeof(".route")+1).compare(".route")!=0) {
		Filename.append(".route");	/* Append .route if it's not already there */
	}

	if (stat(Filename.c_str(),&statdata)==0) {
		Gtk::MessageDialog *Message = new Gtk::MessageDialog("File already exists - overwrite?",
			false,Gtk::MESSAGE_WARNING,Gtk::BUTTONS_OK_CANCEL,true);
		if (Message->run()!=Gtk::RESPONSE_OK) {
			delete Message;
			return;
		}
		delete Message;
	}

	RouteData.Clear();
	try {
		RouteData.Save(Filename);
	}
	catch (Exception Ex) {
		Ex.Show();
	}

	Cfg->SetOpenRouteFile(Filename);
	EditRoute->set_sensitive();
	DrawingArea->Redraw();
}

void MainWindow::on_FileNewWxData_activate() {
	METARs *METARData;
	int Hour;
	Gtk::Window *ProgressWindow;
	Gtk::ProgressBar *ProgressBar,*SubProgressBar;
	Gtk::Button *CancelButton;
	std::list<Volume*>::iterator VolDatai=Data->GetActiveVolume();

	if (VolDatai!=Data->GetVolumes()->end()) {
		try
		{
			std::string gladefile;

			refXml = Gnome::Glade::Xml::create(*GladeFilename, "ProgressWindow");
		}
		catch (const Gnome::Glade::XmlError& ex)
		{
			Gtk::MessageDialog *message = new Gtk::MessageDialog("Error reading data file.  Program cannot continue.",
				false,Gtk::MESSAGE_ERROR,Gtk::BUTTONS_CANCEL,true);
			message->run();
			Gtk::Main::quit();
		}

		refXml->get_widget("ProgressWindow",ProgressWindow);
		refXml->get_widget("SubProgressBar",SubProgressBar);
		refXml->get_widget("ProgressBar",ProgressBar);
		refXml->get_widget("CancelButton",CancelButton);
		ProgressBar->set_text("Downloading");
		SubProgressBar->hide();
		ProgressWindow->show();
		CancelFlag=false;

		ProgressWindow->set_deletable(false);
		CancelButton->signal_clicked().connect(sigc::mem_fun(this,&MainWindow::on_CancelButton_clicked));

		METARData=new METARs(Cfg);
		METARData->Clear();
		for (Hour=0;Hour<Cfg->GetMETARHistory() && !CancelFlag;Hour++) {
			ProgressBar->set_fraction(Hour/(Cfg->GetMETARHistory()+1.0));
			Util::ScheduleMainLoop();
			if (!METARData->DownloadCycleFileURL(METARData->CycleURL(-Hour))) {
				Gtk::MessageDialog *Message = new Gtk::MessageDialog("Download error - continue?",
					false,Gtk::MESSAGE_WARNING,Gtk::BUTTONS_OK_CANCEL,true);
				if (Message->run()!=Gtk::RESPONSE_OK) {
					CancelFlag=true;
				}
				delete Message;
			}
		}
		if (!CancelFlag) {
			ProgressBar->set_text("Processing");
			ProgressBar->set_fraction(Cfg->GetMETARHistory()/(Cfg->GetMETARHistory()+1.0));
			Util::ScheduleMainLoop();

			METARData->CleanUp();
			METARData->Save();
			LinkMETARsToAirports((*VolDatai)->GetAirports(),METARData);
			DrawingArea->Redraw();
		}
		delete METARData;
		delete ProgressWindow;
	} else {
		/* No active data volume */
		Gtk::MessageDialog *Message = new Gtk::MessageDialog("No active data",
			false,Gtk::MESSAGE_WARNING,Gtk::BUTTONS_CANCEL,true);
		Message->run();
		delete Message;
	}
}

void MainWindow::on_FileOpenWxData_activate() {
	METARs *METARData;
	std::list<Volume*>::iterator VolDatai=Data->GetActiveVolume();

	if (VolDatai!=Data->GetVolumes()->end()) {
//TODO Add a progress dialog
		METARData=new METARs(Cfg);
		if (METARData->Load()) {
			LinkMETARsToAirports((*VolDatai)->GetAirports(),METARData);
			DrawingArea->Redraw();
		} else {
			/* Load failed */
//TODO add failure message
		}
		delete METARData;
	} else {
		/* No active data volume */
//TODO add warning notice
	}
}

void MainWindow::on_FileOpenChart_activate() {
	ChartChooserDialog *Dialog=NULL;

	try {
		if (Data->GetActiveVolume()==Data->GetVolumes()->end())
			throw Exception("No active data volume",Exception::LEVEL_ERROR);
		if (Data->GetGeoCharts()->size()==0)
			throw Exception("No charts in active data volume",Exception::LEVEL_ERROR);

		Dialog = new ChartChooserDialog(GladeFilename, Cfg, Data, &RouteData);

		if (Dialog->Run()==Gtk::RESPONSE_OK && Dialog->GetSelectedChart()!=Data->GetGeoCharts()->end()) {
			OpenGeoChart((*Dialog->GetSelectedChart())->GetFileName());
		}
		delete Dialog;
	} catch (Exception Ex) {
		if (Dialog) {
			delete Dialog;
		}
		Ex.Show();
	}
}

void MainWindow::on_FileOpenAircraft_activate() {
	std::string DataDir;
	Gtk::FileChooserDialog FileDialog("Select Aircraft Filename", Gtk::FILE_CHOOSER_ACTION_OPEN);
	DataDir=Cfg->GetUserDataDir();
	Util::ShellExpand(&DataDir);
	FileDialog.set_current_folder(DataDir);

	/* Add filters */
	Gtk::FileFilter AircraftFileFilter;
	AircraftFileFilter.set_name("Aircraft (*.aircraft)");
	AircraftFileFilter.add_pattern("*.aircraft");
	FileDialog.add_filter(AircraftFileFilter);

	Gtk::FileFilter AllFileFilter;
	AllFileFilter.set_name("All Files (*)");
	AllFileFilter.add_pattern("*");
	FileDialog.add_filter(AllFileFilter);

	/* Add response buttons */
	FileDialog.add_button(Gtk::Stock::CANCEL, Gtk::RESPONSE_CANCEL);
	FileDialog.add_button(Gtk::Stock::OPEN, Gtk::RESPONSE_OK);

	if (FileDialog.run()!=Gtk::RESPONSE_OK)
		return;

	if (!AircraftData) {
		AircraftData=new Aircraft();
	}
	try {
		AircraftData->Load(FileDialog.get_filename());
	}
	catch (Exception Ex) {
		Ex.Show();
	}

	Cfg->SetOpenAircraftFile(FileDialog.get_filename());
	EditAircraft->set_sensitive();
	DrawingArea->Redraw();
}

void MainWindow::on_FileOpenRoute_activate() {
	std::string DataDir;
	Gtk::FileChooserDialog FileDialog("Select Route Filename", Gtk::FILE_CHOOSER_ACTION_OPEN);
	DataDir=Cfg->GetUserDataDir();
	Util::ShellExpand(&DataDir);
	FileDialog.set_current_folder(DataDir);

	/* Add filters */
	Gtk::FileFilter RouteFileFilter;
	RouteFileFilter.set_name("Route (*.route)");
	RouteFileFilter.add_pattern("*.route");
	FileDialog.add_filter(RouteFileFilter);

	Gtk::FileFilter AllFileFilter;
	AllFileFilter.set_name("All Files (*)");
	AllFileFilter.add_pattern("*");
	FileDialog.add_filter(AllFileFilter);

	/* Add response buttons */
	FileDialog.add_button(Gtk::Stock::CANCEL, Gtk::RESPONSE_CANCEL);
	FileDialog.add_button(Gtk::Stock::OPEN, Gtk::RESPONSE_OK);

	if (FileDialog.run()!=Gtk::RESPONSE_OK)
		return;

	RouteData.Clear();
	try {
		RouteData.Load(FileDialog.get_filename());
	}
	catch (Exception Ex) {
		Ex.Show();
	}

	Cfg->SetOpenRouteFile(FileDialog.get_filename());
	EditRoute->set_sensitive();
	DrawingArea->Redraw();
}

void MainWindow::on_FileClose_activate() {
	ViewMenu->set_sensitive(false);
	FileClose->set_sensitive(false);
	Cfg->SetOpenChartFile("");
	DrawingArea->CloseChart();
}

void MainWindow::on_FileDataManager_activate() {
	DataManagerDialog *Dialog = new DataManagerDialog(Cfg,GladeFilename, Data, ShapefileData);
	Dialog->Run();

	DrawingArea->SetVolData(*(Data->GetActiveVolume()));
	DrawingArea->SetShapefile(ShapefileData);
	delete Dialog;
}

void MainWindow::on_FileConfigure_activate() {
	ConfigDialog Dialog(GladeFilename,Cfg);
	Dialog.Run();
	if (GPSActive->get_active()) {
		/* Reconnect the GPS Handler in case the update frequency has changed */
		GPSHandlerConnection.disconnect();
		sigc::slot<bool> GPSSlot = sigc::bind(sigc::mem_fun(*this,
              &MainWindow::GPSHandler),0);
		GPSHandlerConnection=Glib::signal_timeout().connect(GPSSlot, Cfg->GetGPSUpdateFrequency());
	}
}

void MainWindow::on_FileFlashlight_activate() {
	FlashlightDialog *Dialog = new FlashlightDialog(GladeFilename);
	Dialog->Run();
	delete Dialog;
	DrawingArea->Redraw();
}

void MainWindow::on_FileQuit_activate() {
	delete GtkWindow;
	Gtk::Main::quit();
}

void MainWindow::on_EditAircraft_activate() {
	AircraftDialog *Dialog = new AircraftDialog(GladeFilename, AircraftData);
	Dialog->Run();

	if (!AircraftData)
		return;

	try {
		AircraftData->Save();
	}
	catch (Exception Ex) {
		Ex.Show();
	}

	delete Dialog;
	DrawingArea->Redraw();
}

void MainWindow::on_EditRoute_activate() {
	RouteDialog *Dialog = new RouteDialog(GladeFilename, Data, &RouteData, AircraftData);
	Dialog->Run();

	try {
		RouteData.Save();
	}
	catch (Exception Ex) {
		Ex.Show();
	}

	delete Dialog;
	DrawingArea->Redraw();
}

void MainWindow::on_HelpAbout_activate() {
	AboutDialog Dialog(GladeFilename);
	Dialog.Run();
}

void MainWindow::on_HelpDisclaimer_activate() {
	DisclaimerDialog Dialog(GladeFilename);
	if (Dialog.Run()!=true) {
		Gtk::Main::quit();
	}
}

void MainWindow::on_ViewZoomIn_activate() {
	DrawingArea->ScaleDisplayScale(1.025);
	DrawingArea->Redraw();
	DrawingArea->PutPositionInStatusbar();
}

void MainWindow::on_ViewZoomOut_activate() {
	DrawingArea->ScaleDisplayScale(0.975);
	DrawingArea->Redraw();
	DrawingArea->PutPositionInStatusbar();
}

void MainWindow::on_ViewZoom30Pct_activate() {
	DrawingArea->SetDisplayScale(0.3);
	DrawingArea->Redraw();
	DrawingArea->PutPositionInStatusbar();
}

void MainWindow::on_ViewZoom60Pct_activate() {
	DrawingArea->SetDisplayScale(0.6);
	DrawingArea->Redraw();
	DrawingArea->PutPositionInStatusbar();
}

void MainWindow::on_ViewZoom100Pct_activate() {
	DrawingArea->SetDisplayScale(1.0);
	DrawingArea->Redraw();
	DrawingArea->PutPositionInStatusbar();
}

void MainWindow::on_ViewZoom150Pct_activate() {
	DrawingArea->SetDisplayScale(1.5);
	DrawingArea->Redraw();
	DrawingArea->PutPositionInStatusbar();
}

void MainWindow::on_ViewZoom200Pct_activate() {
	DrawingArea->SetDisplayScale(2.0);
	DrawingArea->Redraw();
	DrawingArea->PutPositionInStatusbar();
}

void MainWindow::on_ViewZoom400Pct_activate() {
	DrawingArea->SetDisplayScale(4.0);
	DrawingArea->Redraw();
	DrawingArea->PutPositionInStatusbar();
}

void MainWindow::on_ViewPanLeft_activate() {
	DrawingArea->Slide(-0.1,0.0);
	DrawingArea->Redraw();
	DrawingArea->PutPositionInStatusbar();
}

void MainWindow::on_ViewPanRight_activate() {
	DrawingArea->Slide(0.1,0.0);
	DrawingArea->Redraw();
	DrawingArea->PutPositionInStatusbar();
}

void MainWindow::on_ViewPanUp_activate() {
	DrawingArea->Slide(0.0,-0.1);
	DrawingArea->Redraw();
	DrawingArea->PutPositionInStatusbar();
}

void MainWindow::on_ViewPanDown_activate() {
	DrawingArea->Slide(0.0,0.1);
	DrawingArea->Redraw();
	DrawingArea->PutPositionInStatusbar();
}

void MainWindow::on_ViewPanFastLeft_activate() {
	DrawingArea->Slide(-0.8,0.0);
	DrawingArea->Redraw();
	DrawingArea->PutPositionInStatusbar();
}

void MainWindow::on_ViewPanFastRight_activate() {
	DrawingArea->Slide(0.8,0.0);
	DrawingArea->Redraw();
	DrawingArea->PutPositionInStatusbar();
}

void MainWindow::on_ViewPanFastUp_activate() {
	DrawingArea->Slide(0.0,-0.8);
	DrawingArea->Redraw();
	DrawingArea->PutPositionInStatusbar();
}

void MainWindow::on_ViewPanFastDown_activate() {
	DrawingArea->Slide(0.0,0.8);
	DrawingArea->Redraw();
	DrawingArea->PutPositionInStatusbar();
}

void MainWindow::on_ViewGoto_activate() {
	double Lon,Lat;

	GotoDialog Dialog(GladeFilename);
	if (Dialog.Run(&Lon,&Lat)) {
		DrawingArea->CenterOnLonLat(Lon,Lat);
	}
}

void MainWindow::on_ViewFullscreen_activate() {
	if (ViewFullscreen->get_active())
		GtkWindow->fullscreen();
	else
		GtkWindow->unfullscreen();
}

void MainWindow::on_SearchApproaches_activate()
{
	ApproachSearchDialog Dialog(GladeFilename,Data,Cfg);
	Dialog.Run();
}

void MainWindow::on_GPSActive_activate() {
	/* We have to explicitly check the current status before we make changes.  Other parts of
	 * the code may toggle the GPS active status, and we'll just toggle back if we aren't careful.
	 */
	if (GPSActive->get_active()) {
		if (!Gps->GetActive()) {
			SetGPSActive(true);
		}
	} else {
		if (Gps->GetActive()) {
			SetGPSActive(false);
		}
	}
}

void MainWindow::OpenGeoChart(std::string filename) {
	std::string Path;

	if (!Data) {
		throw Exception("Invalid data store",Exception::LEVEL_ERROR);
	}

	if (Data->GetActiveVolume()==Data->GetVolumes()->end()) {
		throw Exception("No active data volume",Exception::LEVEL_ERROR);
	}

	Path=(*Data->GetActiveVolume())->GetDataDir()+"/geocharts/";
	Util::ShellExpand(&Path);
	if (DrawingArea->OpenSectional(Path+filename))
	{
		LastSectionalChange=time(NULL);
		ViewMenu->set_sensitive();
		FileClose->set_sensitive();
		Cfg->SetOpenChartFile(filename);
		GtkWindow->set_title(filename+" - Vifir");
		DrawingArea->SetVolData(*(Data->GetActiveVolume()));
	}
}

void MainWindow::on_GPSStartgpsd_activate() {
	if (Startgpsd()) {
		if (!SetGPSActive(true)) {
			Gtk::MessageDialog *message = new Gtk::MessageDialog("Error connecting to gpsd.  Please check your configuration.",
				false,Gtk::MESSAGE_ERROR,Gtk::BUTTONS_CANCEL,true);
			message->run();
			delete message;
		}
	}

	return;
}

bool MainWindow::GPSHandler(int timer_number) {
	double Lon,Lat,Alt,Trk,Spd;
	double *AltPtr=NULL,*TrkPtr=NULL,*SpdPtr=NULL;
	static bool GPSIdle=false;

	if (Gps->GetFix(&Lon,&Lat,NULL,NULL,NULL)) {
		GPSIdle=false;
		if (GPSAutoSelectChart->get_active() && LastSectionalChange<time(NULL)-10)
		{
			GeoChart *GC;
			GC=Data->GetBestSectional(Lon,Lat,NULL);
			if (GC)
			{
				if (GC->GetFileName().compare(DrawingArea->GetCurrSectionalFileName())!=0)
				{
					OpenGeoChart(GC->GetFileName());
				}
			}
			LastSectionalChange=time(NULL);
		}
		if (Cfg->GetCenterOnGPS())
			DrawingArea->CenterOnLonLat(Lon,Lat);

		if (Gps->GetFix(NULL,NULL,NULL,&Trk,&Spd))
		{
			/* We have track and speed, use them */
			TrkPtr=&Trk;
			SpdPtr=&Spd;
		}
		if (Gps->GetFix(NULL,NULL,&Alt,NULL,NULL))
		{
			/* We have altitude, use it */
			AltPtr=&Alt;
		}
		DrawingArea->MarkLocation(Lon,Lat,TrkPtr,SpdPtr,AltPtr);

	} else {
		if (!GPSIdle) {
			/* No useful GPS data arrived - redraw so we erase the old stuff */
			DrawingArea->Redraw();
			/* No need to continually redraw while there's no new data */
			GPSIdle=true;
		}
	}
	return true;	/* Keep calling this handler */
}

bool MainWindow::SetGPSActive(bool value) {
	if (value) {
		/* Attempt to activate.  If that doesn't work, try to start gpsd
		 * if we're configured that way, then try activating again
		 */
		if (!Gps->Activate()) {
			if (Cfg->GetAutoStartgpsd()) {
				if (!Startgpsd()) {
					return false;
				} else {
					if (!Gps->Activate()) {
						/* Do nothing */
					}
				}
			}
		}
		if (!Gps->GetActive()) {
			GPSActive->set_active(false);
			Gtk::MessageDialog *message = new Gtk::MessageDialog("Error starting GPS.  Is gpsd running?.",
				false,Gtk::MESSAGE_ERROR,Gtk::BUTTONS_CANCEL,true);
			message->run();
			delete message;
			return false;
		}

		/* At this point the GPS is active */
		sigc::slot<bool> GPSSlot = sigc::bind(sigc::mem_fun(*this,
              &MainWindow::GPSHandler),0);
		GPSHandlerConnection=Glib::signal_timeout().connect(GPSSlot, Cfg->GetGPSUpdateFrequency());
		if (GPSActive->get_active()==false) {
			GPSActive->set_active(true);
		}

		/* Inhibit the screen saver and power management if so configured */
		if (Cfg->GetDisableSS())
			InhibitSS();
		if (Cfg->GetDisablePM())
			InhibitPM();
	} else {
		Gps->Deactivate();
		/* At this point the GPS is inactive */
		GPSHandlerConnection.disconnect();
		DrawingArea->ClearLocation();
		DrawingArea->Redraw();
		if (GPSActive->get_active()==true) {
			GPSActive->set_active(false);
		}

		/* UnInhibit the screen saver and power management if so configured */
		if (Cfg->GetDisableSS())
			UnInhibitSS();
		if (Cfg->GetDisablePM())
			UnInhibitPM();
	}
	return true;
}

bool MainWindow::Startgpsd() {
	if (system(Cfg->GetgpsdCmd().c_str())==-1) {
		Gtk::MessageDialog *message = new Gtk::MessageDialog("Error executing gpsd.  Please check your configuration.",
			false,Gtk::MESSAGE_ERROR,Gtk::BUTTONS_CANCEL,true);
		message->run();
		delete message;
		return false;
	}
	sleep(1);	/* Give gpsd a chance to start up */
	return true;
}

void MainWindow::LinkMETARsToAirports(std::vector<Airport*> *Airports, METARs *METARData) {
	std::string LastStation;
	std::vector<Airport*>::iterator Airporti;

	std::vector<Airport*> SortedAirports(*Airports);
	sort(SortedAirports.begin(),SortedAirports.end(),Airport::IndirectLessThan);

	/* Go through each METAR and assign them to an airport */
	Airporti=SortedAirports.begin();
	LastStation="";

	for (std::list<METAR>::iterator METARi=METARData->GetStations()->begin(); METARi!=METARData->GetStations()->end(); ++METARi) {
		if (LastStation.size() && METARi->GetStation().compare(LastStation)==0) {
			/* We're still on the same airport */
			(*Airporti)->AddMETAR(&(*METARi));
		} else {
			/* This is a new airport */
			while ((Airporti!=SortedAirports.end()) && ((*Airporti)->GetICAO().compare(METARi->GetStation())<0)) {
				Airporti++;
			};
			if (Airporti==SortedAirports.end()) {
				break;	/* No more airports to assign METARs to */
			}
			/* Ignore METAR unless it's from an airport we know about */
			if ((*Airporti)->GetICAO().compare(METARi->GetStation())==0) {
				(*Airporti)->ClearMETARs();
				(*Airporti)->AddMETAR(&(*METARi));
			}
			LastStation=METARi->GetStation();
		}
	}
}

bool MainWindow::DBusInit() {
	DBusGConnection *Connection;
	GError *Error=NULL;

	/* Initialize DBus API */
	g_type_init();

	/* Connect to DBus system*/
	Connection=dbus_g_bus_get(DBUS_BUS_SESSION,&Error);
	if (!Connection) {
		/* Ignore errors silently */
		return false;
	}

	/* Get Screen Saver Proxy, ignore any errors */
	DBusInhibitSSProxy=dbus_g_proxy_new_for_name(
		Connection,
		MAINWINDOW_DBUS_SS_SERVICE,
		MAINWINDOW_DBUS_SS_PATH,
		MAINWINDOW_DBUS_SS_INTERFACE);

	/* Get Power Management Proxy, ignore any errors */
	DBusInhibitPMProxy=dbus_g_proxy_new_for_name(
		Connection,
		MAINWINDOW_DBUS_PM_SERVICE,
		MAINWINDOW_DBUS_PM_PATH,
		MAINWINDOW_DBUS_PM_INTERFACE);

	return true;
}

bool MainWindow::DBusShutdown() {
	/* Nothing to do */
	return true;
}

bool MainWindow::InhibitSS() {
	gboolean Result;
	GError *Error=NULL;
	guint32 Cookie;

	if (!DBusInhibitSSProxy || DBusInhibitSSCookie)
		return false;

	DBusInhibitSSCookie=new guint32;
	if (!DBusInhibitSSCookie) {
		/* Ignore errors */
		return false;
	}

	/* Make call */
	Result=dbus_g_proxy_call(
		DBusInhibitSSProxy,
		"Inhibit",
		&Error,
		G_TYPE_STRING,PACKAGE_NAME,			/* Application */
		G_TYPE_STRING,"GPS Active",			/* Reason */
		G_TYPE_INVALID,
		G_TYPE_UINT, &Cookie,			/* Handle to use when UnInhibiting */
		G_TYPE_INVALID);

	if (!Result) {
		/* Ignore errors silently, and give up trying */
		delete DBusInhibitSSCookie;
		DBusInhibitSSCookie=NULL;
		DBusInhibitSSProxy=NULL;
		Cfg->SetDisableSS(false);
		return false;
	}

	*DBusInhibitSSCookie=Cookie;
	return true;
}

bool MainWindow::UnInhibitSS() {
	gboolean Result;
	GError *Error=NULL;

	if (!DBusInhibitSSProxy || !DBusInhibitSSCookie)
		return false;

	/* Make call */
	Result=dbus_g_proxy_call(
		DBusInhibitSSProxy,
		"UnInhibit",
		&Error,
		G_TYPE_UINT, *DBusInhibitSSCookie,
		G_TYPE_INVALID,
		G_TYPE_INVALID);

	delete DBusInhibitSSCookie;
	DBusInhibitSSCookie=NULL;

	if (!Result) {
		/* Ignore errors */
		DBusInhibitSSProxy=NULL;
		Cfg->SetDisableSS(false);
		return false;
	}

	return true;
}

bool MainWindow::InhibitPM() {
	gboolean Result;
	GError *Error=NULL;
	guint32 Cookie;

	if (!DBusInhibitPMProxy || DBusInhibitPMCookie)
		return false;

	DBusInhibitPMCookie=new guint32;
	if (!DBusInhibitPMCookie) {
		/* Ignore errors */
		return false;
	}

	/* Make call */
	Result=dbus_g_proxy_call(
		DBusInhibitPMProxy,
		"Inhibit",
		&Error,
		G_TYPE_STRING,PACKAGE_NAME,			/* Application */
		G_TYPE_STRING,"GPS Active",			/* Reason */
		G_TYPE_INVALID,
		G_TYPE_UINT, &Cookie,			/* Handle to use when UnInhibiting */
		G_TYPE_INVALID);

	if (!Result) {
		/* Ignore errors silently, and give up trying */
		delete DBusInhibitPMCookie;
		DBusInhibitPMCookie=NULL;
		DBusInhibitPMProxy=NULL;
		Cfg->SetDisablePM(false);
		return false;
	}

	*DBusInhibitPMCookie=Cookie;
	return true;
}

bool MainWindow::UnInhibitPM() {
	gboolean Result;
	GError *Error=NULL;

	if (!DBusInhibitPMProxy || !DBusInhibitPMCookie)
		return false;

	/* Make call */
	Result=dbus_g_proxy_call(
		DBusInhibitPMProxy,
		"UnInhibit",
		&Error,
		G_TYPE_UINT, *DBusInhibitPMCookie,
		G_TYPE_INVALID,
		G_TYPE_INVALID);

	delete DBusInhibitPMCookie;
	DBusInhibitPMCookie=NULL;

	if (!Result) {
		/* Ignore errors */
		DBusInhibitPMProxy=NULL;
		Cfg->SetDisablePM(false);
		return false;
	}

	return true;
}

void MainWindow::on_CancelButton_clicked() {
	CancelFlag=true;
}
