#include <iostream>
#include <iomanip>
#include <sstream>
#include <gtkmm/messagedialog.h>
#include <gtkmm/widget.h>
#include <gtkmm/image.h>
#include <gtkmm/stock.h>
#include <gtkmm/main.h>
#include "RouteDialog.h"
#include "WaypointSearchDialog.h"
#include "Geodesic.h"
#include "Util.h"

RouteDialog::RouteDialog(std::string *GladeFilename, DataStore *Data, Route *RouteData, Aircraft *AircraftData)
{
	try
	{
		std::string gladefile;

		refXml = Gnome::Glade::Xml::create(*GladeFilename, "RouteDialog");
	}
	catch (const Gnome::Glade::XmlError& ex)
	{
		Gtk::MessageDialog *message = new Gtk::MessageDialog("Error reading data file.  Program cannot continue.",
			false,Gtk::MESSAGE_ERROR,Gtk::BUTTONS_CANCEL,true);
		message->run();
		Gtk::Main::quit();
	}

	this->GladeFilename=GladeFilename;
	this->Data=Data;
	this->RouteData=RouteData;
	this->AircraftData=AircraftData;

	refXml->get_widget("RouteDialog",GtkDialog);
	refXml->get_widget("CloseButton",CloseButton);
	refXml->get_widget("New",NewButton);
	refXml->get_widget("Reverse",ReverseButton);
	refXml->get_widget("UpButton",UpButton);
	refXml->get_widget("DownButton",DownButton);
	refXml->get_widget("DeleteButton",DeleteButton);
	refXml->get_widget("RouteTreeView",RouteTreeView);
	refXml->get_widget("ScrolledWindow",ScrolledWindow);

	NewButton->signal_clicked().connect(sigc::mem_fun(this,&RouteDialog::on_NewButton_clicked));
	ReverseButton->signal_clicked().connect(sigc::mem_fun(this,&RouteDialog::on_ReverseButton_clicked));
	UpButton->signal_clicked().connect(sigc::mem_fun(this,&RouteDialog::on_UpButton_clicked));
	DownButton->signal_clicked().connect(sigc::mem_fun(this,&RouteDialog::on_DownButton_clicked));
	DeleteButton->signal_clicked().connect(sigc::mem_fun(this,&RouteDialog::on_DeleteButton_clicked));
	CloseButton->signal_clicked().connect(sigc::mem_fun(this,&RouteDialog::on_CloseButton_clicked));
	RouteTreeView->signal_cursor_changed().connect(sigc::mem_fun(this,&RouteDialog::on_TreeView_cursor_changed));

	ListStore = Gtk::ListStore::create(Columns);
	RouteTreeView->set_model(ListStore);
	RouteTreeView->append_column("ID", Columns.ID);
	RouteTreeView->append_column("Leg Dist.", Columns.LegDist);
	RouteTreeView->append_column("Cum. Dist.", Columns.CumDist);
	RouteTreeView->append_column("Leg Time", Columns.LegTime);
	RouteTreeView->append_column("Cum. Time", Columns.CumTime);

	/* Do not allow sorting based on column headers */

	PopulateTreeView(RouteData->GetRoutePoints()->end());	/* No selection */
	UpButton->set_sensitive(false);
	DownButton->set_sensitive(false);
	DeleteButton->set_sensitive(false);

	/* No scrolling at first to set the size.  We enable it when
	 * anything is added
	 */
	ScrolledWindow->set_policy(Gtk::POLICY_NEVER,Gtk::POLICY_NEVER);

	/* Since the size of the data isn't known until runtime, recenter the dialog here */
	GtkDialog->set_position(Gtk::WIN_POS_CENTER_ALWAYS);

	/* If the route is empty, go to add a new route point immediately */
	if (RouteData->GetRoutePoints()->size()==0) {
		on_NewButton_clicked();
	}
}

RouteDialog::~RouteDialog() {
	delete GtkDialog;
}

int RouteDialog::Run() {
	return GtkDialog->run();
}

void RouteDialog::on_NewButton_clicked() {
	std::list<RoutePoint*>::iterator NewRPi;

	WaypointSearchDialog *Dialog = new WaypointSearchDialog(GladeFilename, Data);
	Dialog->Run();
	if (Dialog->GetSelection()) {
		RoutePoint *RP = new RoutePoint(Dialog->GetSelection());
		RouteData->Add(RP);

		for (
			NewRPi=RouteData->GetRoutePoints()->begin();
			NewRPi!=RouteData->GetRoutePoints()->end();
			NewRPi++
			) {
			if ((*NewRPi)==RP) {
				break;
			}
		}

		PopulateTreeView(NewRPi);
	}
	delete Dialog;

	/* Since the size of the data isn't known until runtime, recenter the dialog here */
	ScrolledWindow->set_policy(Gtk::POLICY_NEVER,Gtk::POLICY_AUTOMATIC);
	GtkDialog->set_position(Gtk::WIN_POS_CENTER_ALWAYS);
}

void RouteDialog::on_ReverseButton_clicked() {
	RouteData->Reverse();
	PopulateTreeView(RouteData->GetRoutePoints()->end());
}

void RouteDialog::on_CloseButton_clicked() {
	GtkDialog->response(Gtk::RESPONSE_NONE);
}

void RouteDialog::PopulateTreeView(std::list<RoutePoint*>::iterator Selectedi) {
	std::list<RoutePoint*>::iterator LastPointi;
	double LegDist,CumDist=0.0;
	int SelectedIndex=-1,Count=0;
	std::ostringstream PathOStr;

	/* Remove everything */
	ListStore->clear();

	/* Build from scratch */
	std::list<RoutePoint*> *RoutePoints = RouteData->GetRoutePoints();
	for (std::list<RoutePoint*>::iterator RoutePointi=RoutePoints->begin(); RoutePointi != RoutePoints->end(); RoutePointi++) {
		Gtk::TreeModel::iterator iter = ListStore->append();
		Gtk::TreeModel::Row row = *iter;

		if (RoutePointi==RoutePoints->begin()) {
			LegDist=NAN;
			CumDist=NAN;
		} else {
			LegDist=METERS2NM(Geodesic::GreatCircleDistBear(
				(*RoutePointi)->GetLon(),(*RoutePointi)->GetLat(),
				(*LastPointi)->GetLon(),(*LastPointi)->GetLat()));
			if (isnan(CumDist))
				CumDist=LegDist;
			else
				CumDist+=LegDist;
		}

		AddRoutePoint(
			&row,
			RoutePointi,
			LegDist,
			CumDist,
			RoutePointi==RoutePoints->begin(),
			(*RoutePointi)==RoutePoints->back()
			);
		LastPointi=RoutePointi;

		if (RoutePointi==Selectedi) {
			SelectedIndex=Count;
		}
		Count++;
	}

	if (SelectedIndex>=0) {
		PathOStr<<SelectedIndex;
		RouteTreeView->set_cursor(Gtk::TreePath(PathOStr.str()));
		RouteTreeView->grab_focus();
	}
}

void RouteDialog::AddRoutePoint(Gtk::TreeModel::Row *Row, std::list<RoutePoint*>::iterator RPi, double LegDist, double CumDist, bool First, bool Last) {
	std::ostringstream OText;

	(*Row)[Columns.ID]=(*RPi)->GetID();

	OText.str("");
	if (isnan(LegDist)) {
		OText<<"-";
	} else {
		OText<<(int)LegDist<<"nm";
	}
	(*Row)[Columns.LegDist]=OText.str();

	OText.str("");
	if (isnan(CumDist)) {
		OText<<"-";
	} else {
		OText<<(int)CumDist<<"nm";
	}
	(*Row)[Columns.CumDist]=OText.str();

	OText.str("");
	OText<<std::setiosflags(std::ios::fixed)<<std::setprecision(1);
	if (isnan(LegDist) || !AircraftData) {
		OText<<"-";
	} else {
		float Time=60*LegDist/KPH2KT(AircraftData->GetTASCruise());
		if (Time<=90) {
			OText<<Time<<"m";
		} else {
			OText<<Time/60<<"h";
		}
	}
	(*Row)[Columns.LegTime]=OText.str();

	OText.str("");
	if (isnan(CumDist) || !AircraftData) {
		OText<<"-";
	} else {
		float Time=60*CumDist/KPH2KT(AircraftData->GetTASCruise());
		if (Time<=90) {
			OText<<Time<<"m";
		} else {
			OText<<Time/60<<"h";
		}
	}
	(*Row)[Columns.CumTime]=OText.str();

	(*Row)[Columns.RoutePointi]=RPi;
}

void RouteDialog::on_UpButton_clicked() {
	Glib::RefPtr<Gtk::TreeSelection> RouteSelection = RouteTreeView->get_selection();
	Gtk::TreeModel::iterator iter = RouteSelection->get_selected();
	if (iter) {		/* If anything is selected */
		Gtk::TreeModel::Row Row = *iter;
		std::list<RoutePoint*>::iterator Selectedi;
		std::list<RoutePoint*> *RoutePoints = RouteData->GetRoutePoints();

		for (std::list<RoutePoint*>::iterator RoutePointi=RoutePoints->begin(); RoutePointi != RoutePoints->end(); RoutePointi++) {
			if (RoutePointi==Row[Columns.RoutePointi]) {
				Selectedi=RoutePointi;
			}
		}

		RouteData->Shift(Row[Columns.RoutePointi],-1);

		advance(Selectedi,-1);
		PopulateTreeView(Selectedi);
	}
}

void RouteDialog::on_DownButton_clicked() {
	Glib::RefPtr<Gtk::TreeSelection> RouteSelection = RouteTreeView->get_selection();
	Gtk::TreeModel::iterator iter = RouteSelection->get_selected();
	if (iter) {		/* If anything is selected */
		Gtk::TreeModel::Row Row = *iter;
		std::list<RoutePoint*>::iterator Selectedi;
		std::list<RoutePoint*> *RoutePoints = RouteData->GetRoutePoints();

		for (std::list<RoutePoint*>::iterator RoutePointi=RoutePoints->begin(); RoutePointi != RoutePoints->end(); RoutePointi++) {
			if (RoutePointi==Row[Columns.RoutePointi]) {
				Selectedi=RoutePointi;
			}
		}

		RouteData->Shift(Row[Columns.RoutePointi],1);

		advance(Selectedi,1);
		PopulateTreeView(Selectedi);
	}
}

void RouteDialog::on_DeleteButton_clicked() {
	Glib::RefPtr<Gtk::TreeSelection> RouteSelection = RouteTreeView->get_selection();
	Gtk::TreeModel::iterator iter = RouteSelection->get_selected();
	if (iter) {		/* If anything is selected */
		Gtk::TreeModel::Row Row = *iter;

		RouteData->Del(Row[Columns.RoutePointi]);
		PopulateTreeView(RouteData->GetRoutePoints()->end());	/* No selection */
	}
}

void RouteDialog::on_TreeView_cursor_changed() {
	Glib::RefPtr<Gtk::TreeSelection> RouteSelection = RouteTreeView->get_selection();
	Gtk::TreeModel::iterator iter = RouteSelection->get_selected();
	if (iter) {		/* If anything is selected */
		/* There's probably a more direct way to do this via the ListStore, but I don't
		 * know it.
		 */
		Gtk::TreeModel::Row Row = *iter;

		if ((std::list<RoutePoint*>::iterator)(Row[Columns.RoutePointi])==RouteData->GetRoutePoints()->begin()) {
			/* First one selected */
			UpButton->set_sensitive(false);
			DownButton->set_sensitive(true);
			DeleteButton->set_sensitive(true);
		} else if (*(std::list<RoutePoint*>::iterator)(Row[Columns.RoutePointi])==RouteData->GetRoutePoints()->back()) {
			/* Last one selected */
			UpButton->set_sensitive(true);
			DownButton->set_sensitive(false);
			DeleteButton->set_sensitive(true);
		} else {
			/* Something in the middle selected */
			UpButton->set_sensitive(true);
			DownButton->set_sensitive(true);
			DeleteButton->set_sensitive(true);
		}
	} else {
		UpButton->set_sensitive(false);
		DownButton->set_sensitive(false);
		DeleteButton->set_sensitive(false);
	}
}

RouteDialogModelColumns::RouteDialogModelColumns() {
	add(ID);
	add(LegDist);
	add(CumDist);
	add(LegTime);
	add(CumTime);
	add(RoutePointi);
}
