#include <wordexp.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <cctype>
#include <memory>
#include <gtkmm/main.h>
#include "Util.h"
#include "Exception.h"

void Util::ShellExpand(std::string *str) {
	int result;
	unsigned int i;
	wordexp_t we;

	if ((result=wordexp(str->c_str(),&we,0))==0) {
		(*str)="";
		for (i=0;i<we.we_wordc;i++) {
			if (i!=0)
				(*str)+=" ";
			(*str)+=we.we_wordv[i];
		}
		wordfree(&we);
	}
}

void Util::ChkDir(std::string DirPath) {
	/* Create the directory if it doesn't exist */
	struct stat statdata;
	std::ostringstream Text;

	ShellExpand(&DirPath);
	errno=0;
	if (stat(DirPath.c_str(),&statdata)!=0) {
		if (errno==ENOENT) {
			if (mkdir(DirPath.c_str(),0755)==0) {
				std::cout<<"Created "<<DirPath<<"."<<std::endl;
			} else {
				Text<<"Error creating "<<DirPath<<": "<<strerror(errno)<<" ("<<errno<<")";
				throw Exception(Text.str(),Exception::LEVEL_ERROR);
			}
		}
	}
}

/* Return time based on Greenwich Mean Time */
time_t Util::GMMkTime(struct tm *tm) {
	time_t time;
	char *oldTZ;

	oldTZ=getenv("TZ");
	setenv("TZ","GMT",1);
	tzset();
	time=mktime(tm);

	if (oldTZ)
		setenv("TZ",oldTZ,1);
	else
		unsetenv("TZ");
	tzset();
	return time;
}

bool Util::CopyFile(std::string SrcPath, std::string DestDir, std::string DestName, OverwriteFileDialog *OFD) {
	std::string DstPath,ErrStr;
	std::ifstream ifs;
	std::ofstream ofs;
	struct stat statdata;
	char *buffer=NULL;
	const int bufsize=1024*1024;

	if (DestName.size()) {
		DstPath=DestDir+DestName;
	} else {
		DstPath=DestDir+SrcPath.substr(SrcPath.find_last_of('/')+1,SrcPath.size());
	}

	buffer=(char*)malloc(bufsize);
	if (!buffer) {
		throw Exception("Error allocating memory",Exception::LEVEL_ERROR);
	}

	if (OFD) {
		if (stat(DstPath.c_str(),&statdata)==0) {
			/* File exists */
			switch (OFD->Run(DstPath)) {
				case OverwriteFileDialog::RESPONSE_CANCEL:
					free(buffer);
					throw Exception("Cancelled",Exception::LEVEL_WARNING);
					break;
				case OverwriteFileDialog::RESPONSE_KEEP:
					free(buffer);
					return true;	/* As if we were successful */
				case OverwriteFileDialog::RESPONSE_OVERWRITE:
					break;			/* Fall through as if nothing happened */
			}
		}
	}

	ifs.open(SrcPath.c_str(),std::ios::in);
	if (!ifs.good()) {
		ErrStr=strerror(errno);
		free(buffer);
		throw Exception("Error opening "+SrcPath+": "+ErrStr,Exception::LEVEL_ERROR);
	}

	ofs.open(DstPath.c_str(),std::ios::out);
	if (!ofs.good()) {
		ErrStr=strerror(errno);
		ifs.close();
		free(buffer);
		throw Exception("Error opening "+DstPath+": "+ErrStr,Exception::LEVEL_ERROR);
	}

	do {
		ifs.read(buffer,bufsize);
		ofs.write(buffer,ifs.gcount());
		Util::ScheduleMainLoop();
	} while (ifs.good() && ofs.good());

	if (!ofs.good()) {
		ErrStr=strerror(errno);
		ifs.close();
		ofs.close();
		free(buffer);
		throw Exception("Error writing to "+DstPath+": "+ErrStr,Exception::LEVEL_ERROR);
	}

	if (!ifs.good() && !ifs.eof()) {
		ErrStr=strerror(errno);
		ifs.close();
		ofs.close();
		free(buffer);
		throw Exception("Error reading from "+SrcPath+": "+ErrStr,Exception::LEVEL_ERROR);
	}

	ifs.close();
	ofs.close();
	free(buffer);

	return true;
}

/* Allow the main loop to run (and the GUI to update from time to time */
void Util::ScheduleMainLoop() {
	struct timeval CurrTime;
	double Last,Curr;
	static struct timeval LastTV;
	static bool firsttime=true;

	if (firsttime) {
		LastTV.tv_sec=0;
		LastTV.tv_usec=0;
		firsttime=false;
	}

	Last=LastTV.tv_sec+LastTV.tv_usec/1000000.0;
	gettimeofday(&CurrTime,NULL);
	Curr=CurrTime.tv_sec+CurrTime.tv_usec/1000000.0;
	if (Curr-Last>0.1) {
		while (Gtk::Main::events_pending())
			Gtk::Main::iteration(false);
		gettimeofday(&LastTV,NULL);
	}
}

std::string Util::toUpper(std::string str) {
	for (unsigned int i=0;i<str.length();i++) {
		str[i] = toupper(str[i]);
	}
	return str;
}

std::string Util::DoubleToStr(const double Number) {
	std::ostringstream ostr;

	if (!(ostr<<Number))
		return "";
	return ostr.str();
}

void Util::Explode(const std::string Str, const std::string Delim, std::vector<std::string> *Tokens) {
	size_t Start=0, End=0;

	Tokens->clear();

	End=Str.find_first_of(Delim,Start);
	while (End!=std::string::npos) {
		std::auto_ptr<std::string> Token(new std::string);
		*Token=Str.substr(Start,End-Start);
		Tokens->push_back(*Token);
		Start=End+1;
		End=Str.find_first_of(Delim,Start);
	}
}

/* Convert a glob() pattern into a case-insensitive pattern */
std::string Util::GlobiPattern(std::string Pattern) {
	std::string Result;

	for (unsigned int i=0;i<Pattern.size();i++) {
		if (isalpha(Pattern[i])) {
			Result.append("[");
			Result.append(1,toupper(Pattern[i]));
			Result.append(1,tolower(Pattern[i]));
			Result.append("]");
		} else {
			Result.append(1,Pattern[i]);
		}
	}

	return Result;
}
